# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.13.1
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %%
import matplotlib.pyplot as plt
# %matplotlib inline

# %%
import scopesim
import scopesim_templates 
sim_tp = scopesim_templates


# %%
import astropy
from astropy.io import fits

# %%
import numpy
from pprint import pprint
import copy



# %%
import shutil
import tempfile

import numpy
import astropy.units as u

from micado.config.Environment import Env
from micado.hardware.ShotNoise import ShotNoise
from micado.main.ObservingBlock import ObservingBlock
from micado.hardware.Detector import Detector
from micado.hardware.DetectorArray import DetectorArray
from micado.hardware.DarkCurrent import DarkCurrent
from micado.dataitems.Science import Science
from micado.dataitems.ScienceExtension import ScienceExtension
from micado.templates.MicadoImgObs import MicadoImgObs
from micado.test.functional.dark_test import remove_file_and_links
from micado.util.esokeywords import verify_fitsheaders
from micado.hardware.LinearityCurve import LinearityCurve
from micado.hardware.QuantumEfficiencyCurve import QuantumEfficiencyCurve
from micado.hardware.SurfaceList import SurfaceList
from micado.hardware.Instrument import Instrument
from micado.hardware.RelayOptics import RelayOptics
from micado.hardware.FieldConstantPSF import FieldConstantPSF
from micado.hardware.FilterWheel import FilterWheel
from micado.hardware.SummedExposure import SummedExposure
from micado.hardware.NonCommonPathAberration import NonCommonPathAberration
from micado.hardware.Atmosphere import Atmosphere
from micado.hardware.AtmosphericTERCurve import AtmosphericTERCurve
from micado.hardware.AtmosphericDispersion import AtmosphericDispersion
from micado.hardware.Vibration import Vibration
from micado.hardware.Reflection import Reflection
from micado.hardware.Telescope import Telescope
from micado.hardware.SkycalcTERCurve import SkycalcTERCurve
from micado.dataitems.Detrend import Detrend
from micado.dataitems.DetrendFinal import DetrendFinal
from micado.dataitems.Calibrated import Calibrated
from micado.dataitems.Stack import Stack
from micado.recipes.McdImgDetrend import McdImgDetrend
from micado.recipes.McdImgDetrendFinal import McdImgDetrendFinal
from micado.recipes.McdImgStack import McdImgStack
from micado.recipes.McdImgCalibrate import McdImgCalibrate

from astropy.stats import sigma_clipped_stats
from photutils import datasets
from photutils import DAOStarFinder
import numpy
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm

from astropy.visualization import SqrtStretch
from astropy.visualization.mpl_normalize import ImageNormalize
from photutils import CircularAperture

# %%
my_cluster = sim_tp.stellar.cluster(mass=10000, distance=2000, core_radius=.1)



# %%

# %%
atmo = Atmosphere(
    altitude=3060.,
    longitude=-70.1918,
    latitude=-24.5899,
    temperature=7.,
    humidity=0.1,
    pressure=0.755,
    pwv=2.5,
    # TODO: Enable this again when ScopeSIM 0.4.0 bugs are fixed.
    # skycalc_atmosphere=SkycalcTERCurve(),
)

# %%
telescope = Telescope(
    name="ELT",
    reflection=Reflection(realfilename='LIST_ELT_combined.tbl'),
)

# %%
filt1 = FilterWheel(
    name="Ks",
    minimum_throughput=0.000101,
    outer=0.2,
)
filt2 = FilterWheel(
    name="open",
    minimum_throughput=0.000101,
    outer=0.2,
)

# %%
inst = Instrument(
    name="MICADO",
    ocs__pxscale=0.004,
    plate_scale=0.26666666666,
    filter_wheel_1=filt1,
    filter_wheel_2=filt2,
    static_surfaces=SurfaceList(realfilename="LIST_MICADO_mirrors_static.dat"),
    ncpas_psf=NonCommonPathAberration(realfilename='INST_MICADO_wavefront_error_budget.dat'),
)

# %%
ro = RelayOptics(
    psf=FieldConstantPSF(
        realfilename="PSF_SCAO_ConstPSF_0_5off.fits",
    ),
    surface_list=SurfaceList(
        realfilename="LIST_RO_SCAO_mirrors.dat",
    ),
)

# %%
myexposureaction = SummedExposure()

# Full detector: 4096 pixels
size = 32 * 48

mydetector = Detector(x_size=size, y_size=size)
mydetectorlist = [mydetector]
mydetectorarray = DetectorArray(
    name='',
    detector_array_list=mydetectorlist,
    exposure_action=myexposureaction,
    dark_current=DarkCurrent(value=2.0),
    shot_noise=ShotNoise(state=1),
    qe_curve=QuantumEfficiencyCurve(realfilename="QE_detector_H2RG.dat"),
    linearity=LinearityCurve(realfilename="FPA_linearity.dat"),
)

# A template
science_template = MicadoImgObs(det__dit=1)

# An observingblock
science_ob = ObservingBlock(
    name="MyScienceTest",
    instrument="MICADO",
    target="cluster",
    templates=[science_template],
)


# %%
myextensions = [
    ScienceExtension(detector=det, BITPIX=-64, NAXIS=2)
    for det in mydetectorlist
]

science1 = Science(
    detector=mydetectorarray,
    instrument=inst,
    relay_optics=ro,
    atmosphere=atmo,
    telescope=telescope,
    template=science_template,
    observing_block=science_ob,
    extensions=myextensions,
    
    tpl__expno=0,
)


# %%
science1.scopesim_dictionaries

# %%

# %%
science1.make_simulation()

# %%
data = science1.extensions[0].image
dmin, dmax, dmean, dmed, dstd = data.min(), data.max(), data.mean(), numpy.median(data), data.std()
fig, ax = plt.subplots()
im = ax.imshow(data, norm=LogNorm(vmin=dmed , vmax=dmed + 0.1 * dstd))
fig.colorbar(im)
fig.show()


# %%
pprint(science1.extensions[0].image)

# %%

# %%
pprint(science1.fits_keywords)

# %%

# %% [markdown]
# # Create another science exposure

# %%
myextensions = [
    ScienceExtension(detector=det, BITPIX=-64, NAXIS=2)
    for det in mydetectorlist
]

science3 = Science(
    detector=mydetectorarray,
    instrument=inst,
    relay_optics=ro,
    atmosphere=atmo,
    telescope=telescope,
    template=science_template,
    observing_block=science_ob,
    extensions=myextensions,
    
    tpl__expno=2,
)


# %%
science3.make_simulation()

# %%
data = science3.extensions[0].image
dmin, dmax, dmean, dmed, dstd = data.min(), data.max(), data.mean(), numpy.median(data), data.std()
fig, ax = plt.subplots()
im = ax.imshow(data, norm=LogNorm(vmin=dmed , vmax=dmed + 0.1 * dstd))
fig.colorbar(im)
fig.show()


# %%

# %%
science3.fits_keywords

# %% [markdown]
# # Process the Raw Science into Calibrated Img's

# %%
detrend1 = Detrend(
    raw=science1,
    recipe=McdImgDetrend(),
)
detrend1.make()

# %%
detrend3 = Detrend(
    raw=science3,
    recipe=McdImgDetrend(),
)
detrend3.make()

# %%
detrendfinal1 = DetrendFinal(
    detrend=detrend1,
    recipe=McdImgDetrendFinal(),
)
detrendfinal1.make()


# %%
detrendfinal3 = DetrendFinal(
    detrend=detrend3,
    recipe=McdImgDetrendFinal(),
)
detrendfinal3.make()


# %%
detrendfinal3.extensions[0].image

# %%
calibrateds = [
    Calibrated(
        detrend=detrend,
        recipe=McdImgCalibrate(),
    ) for detrend in [detrendfinal1, detrendfinal3]
]

# %%
for calib in calibrateds:
    calib.make()

# %% [markdown]
# # Create a Stack

# %%
stack = Stack(
    detrends=calibrateds,
    recipe=McdImgStack(),
)

# %%
stack.make()

# %%
data = stack.extensions[0].image
print(data.shape)
dmin, dmax, dmean, dmed, dstd = data.min(), data.max(), data.mean(), numpy.median(data), data.std()
print(dmin, dmax, dmean, dmed, dstd)
fig, ax = plt.subplots(figsize=(14,14))
im = ax.imshow(data, norm=LogNorm(vmin=dmed , vmax=dmed + 0.1 * dstd))
fig.colorbar(im)
fig.show()


# %% [markdown]
# # Source Detection using photutil

# %%
data = stack.extensions[0].image
mean, median, std = sigma_clipped_stats(data, sigma=3.0)
daofind = DAOStarFinder(fwhm=3.0, threshold=5.*std)  
sources = daofind(data - median)  
for col in sources.colnames:  
     sources[col].info.format = '%.8g'  # for consistent table output

print(sources) 

# %%
positions = np.transpose((sources['xcentroid'], sources['ycentroid']))
apertures = CircularAperture(positions, r=4.)
norm = ImageNormalize(stretch=SqrtStretch())

# %%
fig, ax = plt.subplots(figsize=(14,14))
im = ax.imshow(data, norm=LogNorm(vmin=dmed , vmax=dmed + 0.1 * dstd))
apertures.plot(axes=ax, color='red', lw=1.5, alpha=0.5)
fig.colorbar(im)
fig.show()

# %%
