# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.13.1
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# # Recipe Validation using MicadoWISE
#
# A guide on how to use the MicadoWISE, the MICADO Imaging Pipeline Information System as a tool for software validation.

# ## Introduction

# ### Technical preamble
#
# The easiest way to run this notebook is [through Docker](https://gitlab.astro-wise.org/micado/micadowise/-/blob/master/docker/README.md).
#
# The notebook can also be ran in a [standalone MicadoWISE installation](https://gitlab.astro-wise.org/micado/micadowise/-/blob/master/README.md). After installation run
# ```
# conda activate micadowise
# jupyter notebook
# ```
#
# The result should be that a Jupyter Python Notebook interactive user interface opens up in your browser. See this [Kapteyn howto](https://www.astro.rug.nl/KAPTEYN_DB/howto_view.php?stage=2&id=110&auth=0&order=&searchstr=jupyter&personselected=&group=comp) for details.
#
#
# Authors: Gijs A. Verdoes Kleijn (g.a.verdoes.kleijn@rug.nl) and Hugo Buddelmeijer (h.buddelmeijer@rug.nl)

# ### Intended reader audience
#
# We intend that this Jupyter notebook is digestable for a person with the following qualifications:\
#
# * interest in astronomical instrumentation, in particular the MICADO instrument
# * familiar with the Python programming language
# * has executed data calibration and data reduction software pipelines
#
# Please inform the authors if you experience the notebook demands that you possess additional qualifications.

# ### Introduction
#
# The prototype of the MicadoWISE information system can be used to validate requirements on the data processing recipes. This is done by coupling the recipes to the data simulator, processing the simulated raw data, and validating the processed data through their QC parameters. This notebook is a guide through how such a validation mechanism can be implemented.
#
# MicadoWISE is used in the design of the standard imaging data calibration and reduction software for MICADO, the First Light near-IR instrument on the Extremely Large Telescope. It is a digital design in which the core is a machine readable version of the specification of the data items, instrument hardware, software recipes and pipeline. From this we derive a detailed design that is readable for humans. This design documentation, the prototype pipelines and recipes inside it, and data archives are generated in an automated fashion. This approach keeps design and prototype implementation consistent.
#
# MicadoWISE contains python classes for each raw data item type, each processed data type, and each CPL recipe, as well as a model of the instrument and classes for observing templates.
#
# Credits: our approach is based on expertise and lessons learned in our team in earlier projects (e.g. OmegaCAM, MUSE, Euclid).

# ### Imports

# +
# Setup matplotlib.
import matplotlib.pyplot as plt
import matplotlib_inline
# %matplotlib inline
from IPython.display import set_matplotlib_formats
# matplotlib_inline.backend_inline.set_matplotlib_formats('png', 'pdf')

# Import all relevant classes. First a model of the relevant hardware.
from micado.hardware.ShotNoise import ShotNoise
from micado.main.ObservingBlock import ObservingBlock
from micado.hardware.Detector import Detector
from micado.hardware.DetectorArray import DetectorArray
from micado.hardware.DarkCurrent import DarkCurrent
from micado.hardware.SummedExposure import SummedExposure

# The relevant template.
from micado.templates.MicadoImgCalDark import MicadoImgCalDark

# The raw data.
from micado.dataitems.Dark import Dark
from micado.dataitems.Dark import DarkExtension

# The recipes.
from micado.recipes.McdDetDark import McdDetDark

# Processed data.
from micado.dataitems.MasterDark import MasterDark
from micado.dataitems.MasterDark import MasterDarkExtension

# Statistics and requirements.
from micado.main.Imstat import Imstat
from micado.requirements.DPSmasterdark import DPSmasterdark

from micado.toolbox.ditaa_from_wise import ditaa_from_wise

from micado.hardware.LinearityCurve import default_linearitycurve
from micado.hardware.QuantumEfficiencyCurve import default_quantumefficiencycurve
from micado.hardware.ReferencePixelBorder import default_referencepixelborder
# -

# ## Validate the MasterDark recipe.
#
# The following actions are performed in this notebook:
#
# - The mock instrument is setup with a particular known value for the average dark current.
# - Several raw Dark exposures are simulated.
# - A MasterDark is derived using a prototype of the masterdark CPL recipe.
# - Statistics are calculated to verify that requirements on the recipe are fulfilled.
#
# A ASCII-version of a figure from the Data Reduction Library Design shows the steps of the master dark recipe.

print(ditaa_from_wise(MasterDark))

# ### Setup the instrument

# +
# The size of the detector in pixel.
# A small detector is used for this demonstration; the full detector is 4096 pixel.
height_width = 32

# The (high) value of the Dark current.
dark_real = 1. # e/s

# Model a single Detector.
mydetector = Detector(
    x_size=height_width,
    y_size=height_width,
)
mydetectorlist = [mydetector]

# Model the dark current.
mydarkcurrent = DarkCurrent(value=dark_real)

# Model the shotnoise.
myshotnoise = ShotNoise(state=1)

# Model the detector array.
mydetectorarray = DetectorArray(
    name='',
    detector_array_list=mydetectorlist,
    dark_current=mydarkcurrent,
    shot_noise=myshotnoise,
    qe_curve=default_quantumefficiencycurve,
    linearity=default_linearitycurve,
    reference_pixel_border=default_referencepixelborder,
    exposure_action=SummedExposure(state=1),
    default_objects=False,
)
# -

# ### Simulate Several Raw Darks

# +
# Create a template to simulate.
dark_template = MicadoImgCalDark(det__dit=1, det__ndit=1)

# Create an observingblock using this template.
dark_ob = ObservingBlock(
    name="MyDarkTest",
    instrument="MICADO",
    target="empty_sky",
    templates=[dark_template],
)
# -

mydarks = []
for _ in range(10):
    # Create an empty set of pixels for each detector.
    myextensions = [
        DarkExtension(detector=det, BITPIX=-64, NAXIS=2)
        for det in mydetectorlist
    ]
    # Create an empty raw dark.
    mydark = Dark(
        detector=mydetectorarray,
        template=dark_template,
        observing_block=dark_ob,
        extensions=myextensions,
    )
    # Simulate the raw dark.
    mydark.make_simulation()
    
    mydarks.append(mydark)


# Plot one of the darks.
plt.imshow(mydarks[1].extensions[0].image, cmap='gray')
_ = plt.colorbar()

# Plot another dark.
plt.imshow(mydarks[2].extensions[0].image, cmap='gray')
_ = plt.colorbar()

# Print some pixel values.
print(mydarks[2].extensions[0].image)
print(mydarks[2].extensions[0].image.mean(), mydarks[2].extensions[0].image.std())

# The dark average dark current is indeed about 100 e/s, with an expected shotnoise of 10.

# ### Creating a MasterDark

# +
# A recipe. The dark recipe has no parameters.
myrecipe = McdDetDark()

# Create MasterDark.
mymasterdark = MasterDark(
    recipe=myrecipe,
    darks=mydarks,
)
# Use make() or make_cpl() to run the CPL recipe.
mymasterdark.make_proto()
# -

# Inspect pixel values.
print(mymasterdark.extensions[0].image)
print(mymasterdark.extensions[0].uncertainty)

# Inspect QC parameters.
print(f"""{mymasterdark.qc.darkmean = }
{mymasterdark.qc.darkmed = }
{mymasterdark.qc.darkrms = }""")

# The measured dark current in the MasterDark is closer to the true 100 e/s than in the individual raw darks, and the uncertainty has dropped by about a factor of the square root of the number of raw darks. 

# Plot the MasterDark.
plt.imshow(mymasterdark.extensions[0].image, cmap='gray')
_ = plt.colorbar()

# ### Verify MasterDark Requirements
# Use the QC parameters to verify that the MasterDark requirements are met.

myreq = DPSmasterdark()
myreq.precision = 0.05
myreq.dark_true = dark_real
myreq.masterdarkqc = mymasterdark.qc
myreq.make()
if myreq.passed:
    print("Requirement met!")
else:
    print("Requirement not met!")    

# We have verified that the real dark current value can be recovered with a precision of 5%.
