# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.13.1
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %%
import matplotlib.pyplot as plt
# %matplotlib inline

# %%
import scopesim
import scopesim_templates 
sim_tp = scopesim_templates


# %%
import astropy
from astropy.io import fits

# %%
import numpy
from pprint import pprint
import copy



# %%
import shutil
import tempfile

import numpy

from micado.config.Environment import Env
from micado.hardware.ShotNoise import ShotNoise
from micado.main.ObservingBlock import ObservingBlock
from micado.hardware.Detector import Detector
from micado.hardware.DetectorArray import DetectorArray
from micado.hardware.DarkCurrent import DarkCurrent
from micado.dataitems.Science import Science
from micado.dataitems.ScienceExtension import ScienceExtension
from micado.templates.MicadoImgObs import MicadoImgObs
from micado.test.functional.dark_test import remove_file_and_links
from micado.util.esokeywords import verify_fitsheaders
from micado.hardware.LinearityCurve import LinearityCurve
from micado.hardware.QuantumEfficiencyCurve import QuantumEfficiencyCurve
from micado.hardware.SurfaceList import SurfaceList
from micado.hardware.Instrument import Instrument
from micado.hardware.RelayOptics import RelayOptics
from micado.hardware.FieldConstantPSF import FieldConstantPSF
from micado.hardware.FilterWheel import FilterWheel
from micado.hardware.SummedExposure import SummedExposure
from micado.hardware.NonCommonPathAberration import NonCommonPathAberration
from micado.hardware.Atmosphere import Atmosphere
from micado.hardware.AtmosphericTERCurve import AtmosphericTERCurve
from micado.hardware.AtmosphericDispersion import AtmosphericDispersion
from micado.hardware.Vibration import Vibration
from micado.hardware.Reflection import Reflection
from micado.hardware.Telescope import Telescope

# %%
my_cluster = sim_tp.stellar.cluster(mass=10000, distance=2000, core_radius=.1)



# %%

# %%
atmo = Atmosphere(
    altitude=3060.,
    longitude=-70.1918,
    latitude=-24.5899,
    temperature=7.,
    humidity=0.1,
    pressure=0.755,
    pwv=2.5,
    # TODO: Enable this again when ScopeSIM 0.4.0 bugs are fixed.
    # skycalc_atmosphere=SkycalcTERCurve(),
)

# %%
telescope = Telescope(
    name="ELT",
    reflection=Reflection(realfilename='LIST_ELT_combined.tbl'),
)

# %%
filt1 = FilterWheel(
    name="Ks",
    minimum_throughput=0.000101,
    outer=0.2,
)
filt2 = FilterWheel(
    name="open",
    minimum_throughput=0.000101,
    outer=0.2,
)

# %%
inst = Instrument(
    name="MICADO",
    ocs__pxscale=0.004,
    plate_scale=0.26666666666,
    filter_wheel_1=filt1,
    filter_wheel_2=filt2,
    static_surfaces=SurfaceList(realfilename="LIST_MICADO_mirrors_static.dat"),
    ncpas_psf=NonCommonPathAberration(realfilename='INST_MICADO_wavefront_error_budget.dat'),
)

# %%
ro = RelayOptics(
    psf=FieldConstantPSF(
        realfilename="PSF_SCAO_ConstPSF_0_5off.fits",
    ),
    surface_list=SurfaceList(
        realfilename="LIST_RO_SCAO_mirrors.dat",
    ),
)

# %%
myexposureaction = SummedExposure()
#myexposureaction = None

mydetector = Detector()
mydetectorlist = [mydetector]
mydetectorarray = DetectorArray(
    name='testdetector',
    detector_array_list=mydetectorlist,
    exposure_action=myexposureaction,
    dark_current=DarkCurrent(value=2.0),
    shot_noise=ShotNoise(state=1),
    qe_curve=QuantumEfficiencyCurve(realfilename="QE_detector_H2RG.dat"),
    linearity=LinearityCurve(realfilename="FPA_linearity.dat"),
)

# A template
science_template = MicadoImgObs(det__dit=1)

# An observingblock
science_ob = ObservingBlock(
    name="MyScienceTest",
    instrument="MICADO",
    target="cluster",
    templates=[science_template],
)


# %%
myextensions = [
    ScienceExtension(detector=det, BITPIX=-64, NAXIS=2)
    for det in mydetectorlist
]

science1 = Science(
    detector=mydetectorarray,
    instrument=inst,
    relay_optics=ro,
    atmosphere=atmo,
    telescope=telescope,
    template=science_template,
    observing_block=science_ob,
    extensions=myextensions,
)


# %%

# %%
science1.scopesim_dictionaries

# %%
science1.make_simulation()

# %%
from matplotlib.colors import LogNorm
data = science1.extensions[0].image
dmin, dmax, dmean, dmed, dstd = data.min(), data.max(), data.mean(), numpy.median(data), data.std()
fig, ax = plt.subplots()
im = ax.imshow(data, norm=LogNorm(vmin=dmed , vmax=dmed + 0.1 * dstd))
fig.colorbar(im)
fig.show()



# %%
pprint(science1.extensions[0].image)

# %%

# %%
pprint(science1.fits_keywords)

# %%
