"""Create minipipeline.

This script creates darks and flats to test the first three recipes and their
integration into MicadoWISE.
"""

# Import all relevant classes. First a model of the relevant hardware.
from micado.dataitems.DepersistedFlat import DepersistedFlat
from micado.dataitems.Flat import Flat
from micado.dataitems.FlatExtension import FlatExtension
from micado.dataitems.MasterFlat import MasterFlat
from micado.dataitems.Persistence import Persistence
from micado.hardware.ShotNoise import ShotNoise
from micado.main.ObservingBlock import ObservingBlock
from micado.hardware.Detector import Detector
from micado.hardware.DetectorArray import DetectorArray
from micado.hardware.DarkCurrent import DarkCurrent
from micado.hardware.SummedExposure import SummedExposure

# The relevant template.
from micado.recipes.McdImgFlat import McdImgFlat
from micado.templates.MicadoImgCalDark import MicadoImgCalDark

# The raw data.
from micado.dataitems.Dark import Dark
from micado.dataitems.Dark import DarkExtension

# The recipes.
from micado.recipes.McdDetDepersist import McdDetDepersist
from micado.recipes.McdDetDark import McdDetDark

# Processed data.
from micado.dataitems.DepersistedDark import DepersistedDark
from micado.dataitems.MasterDark import MasterDark
from micado.templates.MicadoImgCalFlat import MicadoImgCalFlat
from micado.toolbox.oca_from_python import OCARules


def simulate_minipipeline():
    """Create Darks."""

    # First simulate the hardware.

    # The size of the detector in mm.
    # A small detector is used for this demonstration.
    # The full detector is 30.72 mm.
    height_width = 0.24

    # The (extraordinary high) value of the Dark current in e/s.
    dark_real = 100.0

    # Model a single Detector.
    mydetector = Detector(
        xhw=height_width,
        yhw=height_width,
    )
    mydetectorlist = [mydetector] * 9

    # Model the dark current.
    mydarkcurrent = DarkCurrent(value=dark_real)

    # Model the shotnoise.
    myshotnoise = ShotNoise(state=1)

    # Model the detector array.
    mydetectorarray = DetectorArray(
        name="testdetector",
        detector_array_list=mydetectorlist,
        dark_current=mydarkcurrent,
        shot_noise=myshotnoise,
        exposure_action=SummedExposure(),
    )
    # -

    # ### Simulate Several Raw Darks

    # Create a template to simulate.
    dark_template = MicadoImgCalDark(det__dit=1, det__ndit=1)

    # Create an observingblock using this template.
    dark_ob = ObservingBlock(
        name="MyDarkTest",
        instrument="MICADO",
        target="empty_sky",
        templates=[dark_template],
    )

    # Create Darks and DepersistedDarks
    mydarks = []
    mydepersisteddarks = []
    for _ in range(5):
        # Create an empty set of pixels for each detector.
        myextensions = [
            DarkExtension(detector=det, BITPIX=-64, NAXIS=2) for det in mydetectorlist
        ]
        # Create an empty raw dark.
        mydark = Dark(
            detector=mydetectorarray,
            template=dark_template,
            observing_block=dark_ob,
            extensions=myextensions,
        )
        # Simulate the raw dark.
        mydark.make_simulation()
        mydark.save_fits()

        # Depersist the dark
        mydepersisteddark = DepersistedDark(
            raw=mydark,
            recipe=McdDetDepersist(),
        )
        mydepersisteddark.make_proto()
        mydepersisteddark.save_fits()
        mydepersisteddark.save_sof()

        mydarks.append(mydark)
        mydepersisteddarks.append(mydepersisteddark)

    # ### Creating a MasterDark

    # A recipe. The dark recipe has no parameters.
    myrecipe = McdDetDark()

    # Create MasterDark.
    mymasterdark = MasterDark(
        recipe=myrecipe,
        darks=mydepersisteddarks,
    )
    mymasterdark.make_proto()
    mymasterdark.save_fits()
    mymasterdark.save_sof()

    # Creating Flats
    flat_template = MicadoImgCalFlat(det__dit=1, det__ndit=1)

    # Create an observingblock using this template.
    flat_ob = ObservingBlock(
        name="MyFlatTest",
        instrument="MICADO",
        target="empty_sky",
        templates=[flat_template],
    )

    # Create Darks and DepersistedDarks
    myflats = []
    mydepersistedflats = []
    for _ in range(4):
        # Create an empty set of pixels for each detector.
        myextensions = [
            FlatExtension(detector=det, BITPIX=-64, NAXIS=2) for det in mydetectorlist
        ]
        # Create an empty raw flat.
        myflat = Flat(
            detector=mydetectorarray,
            template=flat_template,
            observing_block=flat_ob,
            extensions=myextensions,
        )
        # Simulate the raw flat.
        myflat.make_simulation()
        myflat.save_fits()

        # Depersist the flat
        mydepersistedflat = DepersistedFlat(
            raw=myflat,
            recipe=McdDetDepersist(),
        )
        mydepersistedflat.make_proto()
        mydepersistedflat.save_fits()
        mydepersistedflat.save_sof()

        myflats.append(myflat)
        mydepersistedflats.append(mydepersistedflat)

    # ### Creating a MasterFlat

    myrecipe = McdImgFlat()

    # Create MasterDark.
    mymasterflat = MasterFlat(
        recipe=myrecipe,
        flats=mydepersistedflats,
        masterdark=mymasterdark,
    )
    mymasterflat.make_proto()
    mymasterflat.save_fits()
    mymasterflat.save_sof()

    # Write OCA rules for this minipipeline.
    allowed = MasterFlat.get_all_input_classes()
    allowed.remove(Persistence)
    delattr(DepersistedDark, "persistence")
    delattr(DepersistedFlat, "persistence")
    rules = OCARules(allowed=allowed)
    rules.add_dataitem(MasterFlat)
    with open("MCDWS_minipipeline.oca", "w") as f:
        f.write(str(rules))


if __name__ == "__main__":
    simulate_minipipeline()
