# -*- coding: utf-8 -*-
"""Hardware that represents a thing. E.g. a Detector."""

from micado.main.Hardware import Hardware


class HardwareThingExtra(Hardware):
    """Hardware that represents a thing. E.g. a Detector.

    This class is necessary to overload scopesim_dictionary.
    """

    @property
    def alias(self):
        """The alias of this HardwareThing, used in ScopeSIM."""
        # TODO: integrate this with
        #       - _convert_normal_keyword in MDBObjectMeta
        #       - real_fits_keywords in MObjectExtra
        #       - hierarch_shortnames in MObject Extra
        # Search for :3 to find them.
        aliasa = self.__class__.__name__.upper().replace("_", "")[:4]
        aliasa = aliasa if aliasa in {"ATMO", "INST", "FILT"} else aliasa[:3]
        return aliasa

    @property
    def scopesim_dictionary(self):
        """Create ScopeSIM dictionary."""
        dd = super(HardwareThingExtra, self).scopesim_dictionary
        dd["alias"] = self.alias
        kwargs = dd.pop("kwargs")
        # Remove parameters like filename, globalname.
        effects = [
            dict(**effect, **{"name": name})
            for name, effect in kwargs.items()
            if isinstance(effect, dict)
            # TODO: Fix PupilWheel Hack
            and effect["class"] not in ["PupilWheel"]
        ]

        from micado.hardware.DetectorArray import DetectorArray
        from micado.hardware.RelayOptics import RelayOptics

        if issubclass(self.__class__, DetectorArray):
            effects.append(
                {
                    "class": "ExtraFitsKeywords",
                    "description": "adds extra FITS keywords from a yaml file",
                    "name": "extra_fits_keywords",
                    "kwargs": {"filename": "FITS_extra_keywords.yaml"},
                }
            )
            effects.append(
                {
                    "class": "SourceDescriptionFitsKeywords",
                    "description": "adds meta data from Source object to FITS header",
                    "name": "source_fits_keywords",
                }
            )
        elif issubclass(self.__class__, RelayOptics):
            effects.append(
                {
                    "class": "ExtraFitsKeywords",
                    "description": "adds extra FITS keywords from a yaml file",
                    "name": "extra_fits_keywords_ro",
                    "kwargs": {"filename": "FITS_extra_keywords_Standalone_RO.yaml"},
                }
            )

        dd["effects"] = effects

        # Props used to be simple, but they can be hierarchical now.
        # Constructing them iteratively is easiest.
        # props = {
        #     k: v
        #     for k, v in kwargs.items()
        #     if not isinstance(v, dict) and k not in ["filename", "globalname"]
        # }
        props = {}
        for k, v in kwargs.items():
            # TODO: Fix REALFILE hack here as well.
            # The "name" of a HardwareThing will overrule the name of any effect.
            if isinstance(v, dict) or k in ["filename", "globalname", "name"]:
                continue
            if "__" in k:
                (kleft, kright) = k.split("__")
                if kleft not in props:
                    props[kleft] = {}
                props[kleft][kright] = v
            else:
                props[k] = v

        # Propagate observation properties to the effect.
        # TODO: find a more generic way to achieve this
        from micado.hardware.DetectorArray import DetectorArray
        from micado.hardware.Atmosphere import Atmosphere
        from micado.hardware.Telescope import Telescope
        from micado.hardware.Instrument import Instrument

        if isinstance(self, DetectorArray):
            props["dit"] = "!OBS.dit"
            props["ndit"] = "!OBS.ndit"

        if isinstance(self, Atmosphere):
            props["airmass"] = "!OBS.airmass"
            props["pupil_angle"] = "!OBS.pupil_angle"
            props["pixel_scale"] = "!INST.pixel_scale"

        if isinstance(self, Telescope):
            props["temperature"] = "!ATMO.temperature"
            props["pixel_scale"] = "!INST.pixel_scale"

        if isinstance(self, Instrument):
            props["pixel_scale"] = props["ocs"]["pxscale"]

        dd["properties"] = props
        return dd
