# -*- coding: utf-8 -*-
"""Templates."""
import re
from typing import Optional, List

import astropy.units as u

from micado.main.MObject import MObject
from micado.util.esokeywords import esokeywords


class TemplateExtra(MObject):
    """A Template."""

    # TODO: Generalize these offsets for different templates.
    offsets = [
        (0 * u.arcsec, 0 * u.arcsec),
        (1 * u.arcsec, 2 * u.arcsec),
        (-1 * u.arcsec, 3 * u.arcsec),
        (-2 * u.arcsec, 1 * u.arcsec),
    ] * 100

    @classmethod
    def name_eso(cls) -> str:
        """Return ESO-compliant name of this Template.

        E.g. MicadoImgCalDark -> MICADO_img_cal_dark.
             MicadoImgObsBgVirtual -> MICADO_img_obs[bg].
        """
        # From https://stackoverflow.com/a/1176023/2097
        n1 = cls.__name__
        n2 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", n1)
        n3 = re.sub("([a-z0-9])([A-Z])", r"\1_\2", n2).lower()
        n4 = n3.split("_")
        if n4[-1] == "virtual":
            n5 = "_".join(n4[:-2]) + f"[{n4[-2]}]"
        else:
            n5 = n3
        return (
            n5.replace("mcd", "MCD")
            .replace("micinsname", "micinsname".upper())
            .replace("micado_", "micado_".upper())
        )

    @classmethod
    def type_eso(cls) -> str:
        """Get ESO type."""
        if "_cal_" in cls.name_eso() or "bg" in cls.name_eso():
            return "Calibration"
        elif "_tec_" in cls.name_eso():
            return "Technical"
        elif "_acq_" in cls.name_eso():
            return "Acquisition"
        elif "_obs" in cls.name_eso() or "_sci_" in cls.name_eso():
            # _sci_ for backwards compatibility with old names.
            return "Science"
        else:
            # Sanity check.
            from micado.main.Template import Template

            assert cls is Template
            return "Unknown"

    @property
    def scopesim_dictionary(self):
        """Create ScopeSIM configuration dictionary."""
        return {
            "object": "observation",
            "alias": "OBS",
            "name": "MICADO_configuration",
            "description": "parameters needed for a MICADO simulation",
            # Packages are needed to find files.
            "packages": ["Armazones", "ELT", "MAORY", "MICADO"],
            "properties": {"dit": self.det__dit, "ndit": self.det__ndit},
        }

    @property
    def obx(self) -> str:
        """OBX file contents for this Template."""
        sparams = [
            f'''{'.'.join(kw.split()[2:]):32}"{value}"'''
            for kw, value in self.fits_keywords.items()
            if " ESO " in kw or " WISE " in kw
        ]
        spar = "\n".join(sparams)
        return f"""TEMPLATE.NAME                   "{self.__class__.name_eso()}"
{spar}
"""

    @classmethod
    def parameters_from_template(cls):
        """Get comma separated list of template parameters.

        Used in rst_from_python and tex_from_python.
        """
        asbad = ["filename", "globalname", "object_id"]
        astemopt = [a for a in cls.get_persistent_properties() if a not in asbad]
        as_ok = [a for a in astemopt if not a.startswith("dpr")]
        sastemplatespace = [a.upper().replace("__", " ") for a in as_ok]
        sastemplate = []
        for sa in sastemplatespace:
            # TODO: Fix yet another place with this MJDOBS MJD-OBS hack.
            if sa == "MJDOBS":
                sa = "MJD-OBS"
            # Fix INS FILT1 NAME
            sa_nonum = sa.replace("1", "").replace("2", "")
            if sa_nonum not in esokeywords:
                sa = "TPL " + sa
                assert sa in esokeywords
            sastemplate.append(sa.replace(" ", "."))

        return ", ".join(sastemplate)
