# -*- coding: utf-8 -*-
"""Tests for Dark.py.

TODO: Add tests for different det_dit and MJDOBS values.
"""

import datetime
import logging
import os
import shutil
import sys
import tempfile
import time

import numpy
import pytest
from astropy.io import fits

from micado.config.Environment import Env
from micado.main.ObservingBlock import ObservingBlock
from micado.dataitems.Dark import Dark
from micado.dataitems.DarkExtension import DarkExtension
from micado.main.Raw import Raw
from micado.templates.MicadoImgCalDark import MicadoImgCalDark
from micado.test.functional.detector_test import generate_detectorarray
from micado.util.esokeywords import verify_fitsheaders


def remove_file_and_links(path):
    """Removes a file and if that file is a link, remove its target too."""
    if not os.path.lexists(path):
        # There is no file.
        return True

    if os.path.islink(path):
        # Follow the link.
        remove_file_and_links(os.readlink(path))

    if os.path.isdir(path):
        # Do not remove directory.
        return False

    if os.path.lexists(path):
        # Remove the actual file.
        try:
            os.remove(path)
        except PermissionError as e:
            # Probably an error on Windows.
            # TODO: Fix this Windows hack.
            print(str(e))

    return not os.path.lexists(path)


def generate_dark(
    dark_real: float = 2.0,
    size: int = 32,
    tpl__start: float = 0.0,
    n_det=1,
):
    """Generate a Dark."""
    mydetectorarray = generate_detectorarray(dark=dark_real, size=size, n_det=n_det)

    # A template
    dark_template = MicadoImgCalDark(det__dit=1, det__ndit=1)

    # An observingblock
    dark_ob = ObservingBlock(
        name="MyDarkTest",
        instrument="MICADO",
        target="empty_sky",
        templates=[dark_template],
    )

    # DarkExtensions
    myextensions = [
        DarkExtension(detector=det, BITPIX=-64, NAXIS=2)
        for det in mydetectorarray.detector_array_list
    ]

    dark1 = Dark(
        detector=mydetectorarray,
        template=dark_template,
        observing_block=dark_ob,
        tpl__start=tpl__start,
        extensions=myextensions,
    )
    assert dark1.detector == mydetectorarray
    dark1.make_simulation()
    return dark1


def compare_raws(raw1: Raw, raw2: Raw):
    """Compare two Raws."""
    assert numpy.all(raw2.extensions[0].image == raw1.extensions[0].image)
    assert raw1.extensions[0].NAXIS1 == raw2.extensions[0].NAXIS1
    assert raw1.extensions[0].NAXIS2 == raw2.extensions[0].NAXIS2
    assert raw1.detector == raw2.detector
    assert raw1.template == raw2.template
    assert raw1.observing_block == raw2.observing_block
    ext1 = raw1.extensions[0]
    ext2 = raw2.extensions[0]
    assert ext1.detector == ext2.detector
    assert ext1 == ext2, ext1.get_differences(ext2)
    assert raw2 == raw1


def test_dark():
    """Test creation of a Dark."""
    # Set the real dark current and create a DetectorArray.
    dark_real = 10000.0

    dark1 = generate_dark(dark_real=dark_real)
    dark1.save_fits()
    fn = dark1.filename
    verify_fitsheaders(fn)

    dark2 = Dark(filename=fn)
    compare_raws(dark1, dark2)

    fn3 = tempfile.NamedTemporaryFile(suffix=".fits", delete=False).name
    shutil.copy(fn, fn3)

    dark3 = Dark(filename=fn3)
    dark3.unique_filename()

    hash1 = dark1.filename.split("_")[-1]
    hash3 = dark3.filename.split("_")[-1]
    assert hash1 == hash3, "Bad file hashes"

    remove_file_and_links(dark1.filename)
    remove_file_and_links(dark3.filename)
    remove_file_and_links(fn3)

    # Save with self.filename set.
    dark3.save_fits()
    remove_file_and_links(dark3.filename)

    # Save while explicitly setting a filename.
    dark3.save_fits(filename=dark3.filename)
    remove_file_and_links(dark3.filename)


def test_dark_multi():
    """Test creation of a Dark."""
    # Set the real dark current and create a DetectorArray.
    dark_real = 10000.0
    mydetectorarray = generate_detectorarray(dark=dark_real, size=32, n_det=2)

    # A template
    dark_template = MicadoImgCalDark(det__dit=1)

    # An observingblock
    dark_ob = ObservingBlock(
        name="MyDarkTest",
        instrument="MICADO",
        target="empty_sky",
        templates=[dark_template],
    )

    mydarkextensions = [
        DarkExtension(detector=mydetector, BITPIX=-64, NAXIS=2)
        for mydetector in mydetectorarray.detector_array_list
    ]

    for myext in mydarkextensions:
        myext.image = numpy.ones(
            (myext.NAXIS1, myext.NAXIS2),
            dtype=numpy.float64,
        )

    dark1 = Dark(
        detector=mydetectorarray,
        template=dark_template,
        observing_block=dark_ob,
        extensions=mydarkextensions,
    )

    from pprint import pprint

    pprint(dark1.fits_keywords)
    dark1.save_fits()
    fn = dark1.filename

    dark2 = Dark(filename=fn)

    assert dark2.detector.dark_current.value == dark1.detector.dark_current.value
    assert dark2.detector == dark1.detector
    assert dark2.extensions[0] == dark1.extensions[0], dark2.extensions[
        0
    ].get_differences(dark1.extensions[0])
    assert dark2.extensions[1] == dark1.extensions[1]
    assert dark2 == dark1

    remove_file_and_links(dark1.filename)


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Test will fail on Windows")
def test_template():
    """Test reading a template from FITS."""

    dark_template = MicadoImgCalDark(det__dit=1)

    with tempfile.NamedTemporaryFile() as ftemplate:
        # TODO: make this possible:
        #   dark_template.save_fits()
        header = fits.Header(dark_template.fits_keywords)
        hdu = fits.PrimaryHDU(header=header)
        hdus = fits.HDUList([hdu])
        hdus.writeto(ftemplate)

        with pytest.raises(NotImplementedError):
            template2 = MicadoImgCalDark(filename=ftemplate.name)
            assert dark_template == template2


def test_resave():
    """Test saving a dark twice."""
    dark1 = generate_dark()
    dark1.save_fits()
    dark2 = Dark(filename=dark1.filename)
    # save original dark again
    dark1.save_fits()
    dark3 = Dark(filename=dark1.filename)
    assert dark3 == dark2
    assert dark2 == dark1


def test_random_seed():
    """Test whether setting the random seed works."""
    now = datetime.datetime.now()

    dark1 = generate_dark()
    dark2 = generate_dark()
    dark3 = generate_dark(tpl__start=now.timestamp())
    dark3.save_fits()
    # Sleep two seconds to ensure that there is no hidden time dependency,
    # e.g. the DATE keyword.
    time.sleep(2)
    dark4 = generate_dark(tpl__start=now.timestamp())
    # Saving to file is necessary to properly compare the headers.
    dark4.save_fits()
    assert not (dark1.extensions[0].image == dark2.extensions[0].image).all()
    assert (dark3.extensions[0].image == dark4.extensions[0].image).all()
    assert dark4 == dark3
    for fn in [dark3.filename, dark4.filename]:
        if os.path.exists(fn):
            os.remove(fn)
