from datetime import datetime

from common.database.Context import context
from micado.dataitems.Flat import Flat
from micado.dataitems.FlatExtension import FlatExtension
from micado.main.ObservingBlock import ObservingBlock
from micado.templates.MicadoImgCalFlat import MicadoImgCalFlat
from micado.test.functional.detector_test import generate_detectorarray
from micado.test.functional.science_test import get_filt_inst_ro
from micado.util.Processor import LocalProcessor
from micado.util.SimulateRawJob import SimulateRawFlatJob


def test_flat_recipe():
    """Test the creation of Flat through the lpu/dpu recipe."""

    # Hardware and observation setup.
    now = datetime.now()

    # Detector array
    detectorarray = generate_detectorarray()

    # A template
    template = MicadoImgCalFlat()

    # An observingblock
    observing_block = ObservingBlock(
        target=SimulateRawFlatJob.observingblock_target,
        templates=[template],
        instrument="MICADO",
    )

    _, _, instrument, relay_optics = get_filt_inst_ro()

    # Generate a flat here.
    # FlatExtensions, this is probably something that should be moved to the make method though
    extensions = [
        FlatExtension(detector=det, BITPIX=-64, NAXIS=2)
        for det in detectorarray.detector_array_list
    ]

    detectorarray.commit()
    template.commit()
    observing_block.commit()
    instrument.commit(),
    relay_optics.commit()

    # Generate a flat through the recipe.
    lpu = LocalProcessor()
    lpu.run(
        "rawflat",
        commit=True,
        detector=detectorarray,
        instrument=instrument,
        mjdobs=now.isoformat(),
        observing_block=observing_block,
        relay_optics=relay_optics,
        skip_store=True,
        template=template,
        tpl__start=now.isoformat(),
    )

    object_through_lpu = lpu.get_job_result()[0]

    object_manually = Flat(
        detector=detectorarray,
        extensions=extensions,
        instrument=instrument,
        mjdobs=now.timestamp(),
        observing_block=observing_block,
        relay_optics=relay_optics,
        template=template,
        tpl__start=now.timestamp(),
    )
    object_manually.make()
    # Saving the FITS file might overwrite the file created on the DPU. However,
    # this should not matter because the filename contains a hash of the
    # contents.
    object_manually.save_fits()

    # Tests
    assert object_through_lpu.observing_block is observing_block
    assert object_through_lpu.template is template
    assert object_through_lpu.detector is detectorarray

    # The tpl__start should ensure all the pixel values are identical.
    assert object_through_lpu == object_manually, object_through_lpu.get_differences(
        object_manually
    )

    # Cleanup.
    # Delete test objects from database.
    to_delete = [
        object_manually,
        object_through_lpu,
        detectorarray,
        observing_block,
        template,
        instrument,
        relay_optics,
        *object_manually.extensions,
        *object_through_lpu.extensions,
        *detectorarray.detector_array_list,
    ]
    for obj in to_delete:
        if obj.persists():
            # TODO: Only delete objects that were actually created during this test.
            context.delete(obj, commit=True)
