# -*- coding: utf-8 -*-
"""Generate an association matrix."""

import itertools
import random
from collections import OrderedDict
from dataclasses import dataclass
from functools import lru_cache
from typing import Dict, Any, List, Optional

from micado.dataitems.CalibHCI import CalibHCI
from micado.dataitems.CalibratedCatalog import CalibratedCatalog
from micado.dataitems.CoroOffset import CoroOffset
from micado.dataitems.CubeHCI import CubeHCI
from micado.dataitems.Detrend import Detrend
from micado.dataitems.DetrendBase import DetrendBase
from micado.dataitems.DetrendBkgfield import DetrendBkgfield
from micado.dataitems.DetrendFinalAi import DetrendFinalAi
from micado.dataitems.DetrendFinalBase import DetrendFinalBase
from micado.dataitems.DetrendFinalHCIBase import DetrendFinalHCIBase
from micado.dataitems.DetrendFinalHCIScienceBase import DetrendFinalHCIScienceBase
from micado.dataitems.DetrendFinalSiaiBase import DetrendFinalSiaiBase
from micado.dataitems.DetrendHCIBase import DetrendHCIBase
from micado.dataitems.DistortionBase import DistortionBase
from micado.dataitems.DistortionRefBase import DistortionRefBase
from micado.dataitems.DistortionWam import DistortionWam
from micado.dataitems.Nonlinearity import Nonlinearity
from micado.dataitems.PhotomBase import PhotomBase
from micado.dataitems.PhotomStdfield import PhotomStdfield
from micado.dataitems.DpsicsStaroffsetOpt import DpsicsStaroffsetOpt
from micado.dataitems.Regrid import Regrid
from micado.dataitems.Science import Science
from micado.dataitems.StarcenterBase import StarcenterBase
from micado.dataitems.StarcenterPupil import StarcenterPupil
from micado.dataitems.StarcenterWaffle import StarcenterWaffle
from micado.dataitems.Staroffset import Staroffset
from micado.dataitems.Stdfield import Stdfield
from micado.main.Opt import Opt
from micado.main.RecipeStepExtra import RecipeStepBase
from micado.main.Sky import Sky
from micado.main.SkySiaiBase import SkySiaiBase
from micado.recipes.ApplyAstrom import ApplyAstrom
from micado.recipes.ApplyMasterDark import ApplyMasterDark
from micado.recipes.ApplyNonlinearity import ApplyNonlinearity
from micado.recipes.ApplyPersistence import ApplyPersistence
from micado.recipes.ComputePhotomInScienceFilter import ComputePhotomInScienceFilter
from micado.recipes.CorrectGeomDistortion import CorrectGeomDistortion
from micado.recipes.CreateSourceCatalogFinal import CreateSourceCatalogFinal
from micado.recipes.DeriveDpsicsDetrend import DeriveDpsicsDetrend
from micado.toolbox.containedint import ContainedInt
from micado.dataitems.Astrom import Astrom
from micado.dataitems.Background import Background
from micado.dataitems.Calibrated import Calibrated
from micado.dataitems.DetrendFinal import DetrendFinal
from micado.dataitems.IllumCorr import IllumCorr
from micado.dataitems.MasterDark import MasterDark
from micado.dataitems.MasterFlat import MasterFlat
from micado.dataitems.Stack import Stack
from micado.main.DataItem import DataItem
from micado.toolbox.ditaa_from_wise import (
    get_ditaaclass,
    text_from_template,
    text_from_dataitem,
    get_applicable_templates_from_dataitems,
    get_raws_derived,
)
from micado.toolbox.ditaacreator import (
    ProcessMini,
    PosArt,
    RecipeShort,
    Canvas,
    Line,
    Process,
    Template as TemplateDitaa,
    DataExternal,
    DataStatic,
)
from micado.util.classlists import (
    CLASSES_HIGH_CONTRAST_IMAGING,
    CLASSES_HIGH_CONTRAST_IMAGING_ALL,
    CLASSES_HIGH_CONTRAST_IMAGING_FOCAL,
    CLASSES_HIGH_CONTRAST_IMAGING_FOCAL_ALL,
    CLASSES_HIGH_CONTRAST_IMAGING_PUPIL,
    CLASSES_HIGH_CONTRAST_IMAGING_PUPIL_ALL,
    CLASSES_STANDARD_IMAGING,
    CLASSES_STANDARD_IMAGING_FINAL,
    CLASSES_STANDARD_IMAGING_ALL,
    CLASSES_ASTROMETRIC_IMAGING,
    CLASSES_ASTROMETRIC_IMAGING_FINAL,
    CLASSES_ASTROMETRIC_IMAGING_ALL,
    CLASSES_STANDARD_IMAGING_SMALL,
    CLASSES_HIGH_CONTRAST_IMAGING_SMALL,
    CLASSES_ASTROMETRIC_IMAGING_SMALL,
    CLASSES_UPTO_REDUCED,
    CLASSES_UPTO_REDUCED_FINAL,
    CLASSES_UPTO_REDUCED_ALL,
)
from micado.util.paths import PATH_DOCS_GENERATED

DASHED_HORIZONTAL = False


@dataclass
class NeighbourArt:
    """Art that is a neighbour of other Art."""

    art: "ConnPosArt" = None
    distance: int = None
    line: Optional[str] = "|"


@dataclass
class ConnPosArt(PosArt):
    """Connected Position Art."""

    centered: bool = True
    up: NeighbourArt = None
    down: NeighbourArt = None
    left: NeighbourArt = None
    right: NeighbourArt = None

    @property
    def tension(self):
        """The tension felt by this PosArt."""
        tension = 0
        if self.leftp < 0:
            tension += 1000
        if self.top < 0:
            tension += 1000
        if self.top > 0:
            tension += 0.01
        if self.up is not None:
            # Prefer to have up two above us.
            if self.up.distance is not None:
                tension += (
                    2
                    / (1 + self.top / 100)
                    * abs(self.top - self.up.art.bottom - self.up.distance)
                )
            # And centered
            tension += 0.5 * abs(self.center_horizontal - self.up.art.center_horizontal)
        if self.left is not None:
            # Prefer to have left two to the left
            if self.left.distance is not None:
                tension += (
                    2
                    / (1 + self.leftp / 100)
                    * abs(self.leftp - self.left.art.rightp - self.left.distance)
                )
            # And centered
            tension += 3.5 * abs(self.center_vertical - self.left.art.center_vertical)
        if self.down is not None:
            if self.down.distance is not None:
                tension += (
                    1
                    / (1 + self.top / 100)
                    * abs(self.bottom - self.down.art.top + self.down.distance)
                )
            # And centered
            tension += abs(self.center_horizontal - self.down.art.center_horizontal)
        if self.right is not None:
            # Prefer to have left two to the left
            if self.right.distance is not None:
                tension += (
                    1
                    / (1 + self.leftp / 100)
                    * abs(self.rightp - self.right.art.leftp + self.right.distance)
                )
            # And centered
            tension += 2.5 * abs(self.center_vertical - self.right.art.center_vertical)

        return tension

    @property
    def tension_left(self):
        """Cumulative tension to the left.

        Currently unused.
        """
        rr = self
        ll = self.left
        ww = 0
        tens = 0
        while ll is not None:
            tens += abs((rr.leftp - ll.art.rightp) - (ll.distance + ww))
            ww += ll.distance + rr.art.width
            rr = ll.art
            ll = ll.art.left
        return tens


@dataclass
class ConnCanvas(Canvas):
    """Canvas with connected PosArts."""

    @property
    def tension(self):
        """The tension felt by the arts."""
        tension = sum(po.tension for po in self.arts if isinstance(po, ConnPosArt))

        for a1, a2 in itertools.permutations(self.arts, 2):
            # https://stackoverflow.com/questions/27152904/calculate-overlapped-area-between-two-rectangles
            dx = min(a1.rightp, a2.rightp) - max(a1.x, a2.x)
            dy = min(a1.bottom, a2.bottom) - max(a1.y, a2.y)
            if dx >= 0 and dy >= 0:
                tension += 4 * dx * dy

        return tension

    def shuffle(self):
        """Shuffle the Canvas.

        This algorithm does not work with ContainedInts, because the
        backing up and restoring of the position only works with normal ints.
        TODO: Ensure this algorithm works with ContainedInts.
        """

        def move_l(pob):
            """Move left."""
            pob.x -= 1

        def move_r(pob):
            """Move right."""
            pob.x += 1

        def move_u(pob):
            """Move up."""
            pob.y -= 1

        def move_d(pob):
            """Move down."""
            pob.y += 1

        moves = [move_l, move_r, move_u, move_d]
        arts_shuffled = random.sample(self.arts, len(self.arts))
        too = self.tension
        po_same, move_same = None, None
        for po in arts_shuffled:
            xo, yo, to = po.x, po.y, self.tension
            for move in random.sample(moves, len(moves)):
                move(po)
                if self.tension < to:
                    break
                if self.tension == to:
                    po_same, move_same = po, move
                    po.x, po.y = xo, yo
                else:
                    po.x, po.y = xo, yo
            if self.tension < too:
                break
        else:
            # No move led to a lower tension
            if move_same is not None:
                move_same(po_same)


@lru_cache(maxsize=1000000)
def can_zig_zag(l1: str, l2: str):
    """Can we zigzag through these two lines?

    Args:
        l1: line 1
        l2: line 2

    True:
      '   xx   '
      'xx     x'
    False:
      '   xx   '
      'xxx    x'
    """
    # first step is free
    lsplus = [" " + li + " " for li in [l1, l2]]
    scores = [
        [
            cm.isspace() * (not cl.isspace())
            + 2 * (not cm.isspace())
            + cm.isspace() * (not cr.isspace())
            for cl, cm, cr in zip(li[:-2], li[1:-1], li[2:])
        ]
        for li in lsplus
    ]
    sumscores = [sum(li) for li in zip(*scores)]
    goods = [li < 3 for li in sumscores]

    return all(goods)


def get_bases(cc):
    """Get base classes recursively."""
    if issubclass(cc, DataItem) and "dataitems" in cc.__module__:
        yield cc
        for cci in cc.__bases__:
            yield from get_bases(cci)


class AssociationMatrix:
    """The full AssociationMatrix as DITAA ASCII art."""

    def __init__(self, classes_to_draw, classes_final, classes_all, big=False):
        self.classes_to_draw = classes_to_draw
        # The classes that should be drawn.

        self.classes_final = classes_final
        # The classes that are the final products of the pipeline.

        self.classes_all = classes_all
        # All classes that are allowed to show up in the matrix.

        self.big = big
        # Whether this is a big matrix with DRL functions included.

        self.y_step = 0
        # The y-coordinate of the step currently being added; transient.

        self.thearts = OrderedDict()
        # Dictionary with ASCII art to draw.

        self.conncanvas: ConnCanvas = None
        # The actual Canvas with the ASCII art on it.

        self.legend = False
        # Include legend or not

        self.contained_ints_dict = {
            "ypos_template": ContainedInt(10),
            "ypos_raw": ContainedInt(30),
            "ypos_recipe": ContainedInt(50),
            "xpos_0": ContainedInt(15),
        }
        # Dictionary with coordinates, so boxes can be moved together.

        self.rsteps_duoinput = []

    def shuffle(self, cc: ConnCanvas):
        """Shuffle ConnCanvas cc by permutating contained ints."""
        contained_ints = [cf for cf in self.contained_ints_dict.values()]
        cfs_shuffled = random.sample(contained_ints, len(contained_ints))
        too = cc.tension
        move_same = None
        for cf in cfs_shuffled:
            to = cc.tension
            moves = [object.__getattribute__(cf, mm) for mm in ["__iadd__", "__isub__"]]
            for move in random.sample(moves, len(moves)):
                move(1)
                if cc.tension < to:
                    break
                if cc.tension == to:
                    move_same = move
                    move(-1)
                else:
                    move(-1)
            if cc.tension < too:
                break
        else:
            # No move led to a lower tension
            if move_same is not None:
                move_same(1)

    def remove_line(self, lines: List[str]):
        """If there are two consecutive empty lines in the canvas, remove one."""
        for i, (l1, l2, l3) in enumerate(zip(lines[2:], lines[1:-1], lines[:-2])):
            if l2.isspace() and can_zig_zag(l1, l3):
                break
        else:
            return False

        ypos_to_remove = i + 1
        cis_to_move: List[ContainedInt] = [
            ci
            for k, ci in self.contained_ints_dict.items()
            if "ypos" in k and ci > ypos_to_remove
        ]
        for ci in cis_to_move:  # NOSONAR
            ci -= 1  # NOSONAR
        return lines[: i + 1] + lines[i + 2 :]

    def get_duosteps(self):
        """Get those recipesteps that require two inputs."""
        recipesteps = set()
        for do in self.classes_to_draw:
            recipestep = do.recipe.prop_type.process.prop_type
            while recipestep is not None:
                recipesteps.add(recipestep)
                recipestep = recipestep.get_previous_step()
        rsteps_multiinput = [
            r for r in recipesteps if len(r.get_other_inputdata_full()) > 1
        ]
        # 4 used to be CorrectGeomDistortion from Astrom, now it is 2 as well
        # Their extra input is from a different imaging pipeline, so it does not
        # have to be in the association matrix.
        # ComputePhotomInScienceFilter has two PHOTOM_TAB's as input.
        # TODO: Fix HCI online, that is, with DpsicsStaroffsetOpt
        if DpsicsStaroffsetOpt not in self.classes_to_draw:
            assert {len(r.get_other_inputdata_full()) for r in rsteps_multiinput} <= {2}
        self.rsteps_duoinput = [
            r
            for r in recipesteps
            if len(r.get_other_inputdata_full()) == 2
            and r not in {ComputePhotomInScienceFilter}
        ]
        assert {r.get_previous_step() for r in self.rsteps_duoinput} <= {
            None,
            DeriveDpsicsDetrend,
            # TODO: Does having CreateSourceCatalogFinal here make sense?
            CreateSourceCatalogFinal,
        }
        allowed_other_input_for_duoinput = {
            # SI, HCI
            Detrend,
            DetrendBase,
            DistortionWam,
            # SI
            DetrendFinalBase,
            # AI
            DetrendFinalAi,
            DetrendFinalSiaiBase,
            # QACITS
            MasterDark,
            # "QACITSDETRENDSKP_IMG",
        }
        actual_other_input_for_duoinput = {
            r.get_other_inputdata_full()[0] for r in self.rsteps_duoinput
        }
        assert actual_other_input_for_duoinput <= allowed_other_input_for_duoinput

    def get_lines_up_left(self):
        """Get lines up and left."""
        lines = []
        po: ConnPosArt
        for po in self.conncanvas.arts:
            if po.up is not None and po.up.line:
                lines.append(
                    ConnPosArt(
                        x=po.x - 1,
                        y=po.up.art.bottom - 1,
                        centered=False,
                        art=Line(
                            vertical=True,
                            distance=po.top - po.up.art.bottom + 2,
                            style=po.up.line,
                            connector="3",
                        ),
                    )
                )
            if po.left is not None and po.left.line:
                lines.append(
                    ConnPosArt(
                        x=po.left.art.rightp - 1,
                        y=po.y - 1,
                        centered=False,
                        art=Line(
                            vertical=False,
                            distance=po.leftp - po.left.art.rightp + 2,
                            style=po.left.line,
                            connector="4",
                        ),
                    )
                )
        return lines

    def get_duo_lines(self):
        """Get duo lines."""
        lines = []
        for rs in self.rsteps_duoinput:
            do_input = rs.get_other_inputdata_full()[0]
            # TODO: Fix hack for ApplyAstrom.
            if rs == ApplyAstrom:
                do_input = DetrendFinalBase
            name_input = do_input.name_eso()
            name_rs = rs.name_eso()
            # There are quite some exceptions to this rule, so only add the
            # duo lines if the first input is in thearts.
            if name_input not in self.thearts:
                assert rs != ApplyMasterDark
                continue
            assert name_rs in self.thearts
            for art in self.thearts[name_rs]:
                # Create horizontal lines.
                lines.append(
                    ConnPosArt(
                        x=self.thearts[name_input].rightp - 1,
                        y=self.thearts[name_input].y - 1,
                        centered=False,
                        art=Line(
                            vertical=False,
                            distance=art.x - self.thearts[name_input].rightp + 1,
                            connector="5",
                            style=":" if DASHED_HORIZONTAL else "|",
                        ),
                    )
                )
                # Vertical lines.
                lines.append(
                    ConnPosArt(
                        x=art.x - 1,
                        y=self.thearts[name_input].y - 1,
                        centered=False,
                        art=Line(
                            vertical=True,
                            distance=art.y
                            - self.thearts[name_input].y
                            + (-1 if self.big else 1),
                            connector="6",
                        ),
                    )
                )
        # TODO: This function does nothing when big==False except make 1 dashed
        #  line.
        return lines

    def is_point_in_arts(self, x, y):
        """Is the point at x, y within the arts?"""
        # Some arts are actually lists of arts, so flatten.
        arts2 = [a if isinstance(a, list) else [a] for a in self.thearts.values()]
        arts3 = [a for aa in arts2 for a in aa]
        is_ins = [a.is_inside(x, y) for a in arts3]
        return any(is_ins)

    def add_lines(self):
        """Add lines to connect everything."""
        lines = self.get_lines_up_left() + self.get_duo_lines()

        for cpo in lines:
            assert isinstance(cpo.art, Line)
        line_ends = [
            ConnPosArt(
                x=cpo.x,
                y=cpo.y,
                centered=cpo.centered,
                art=Line(
                    vertical=cpo.art.vertical,
                    distance=0,
                    connector="1",
                ),
            )
            for cpo in lines
            if not self.is_point_in_arts(
                x=cpo.x,
                y=cpo.y,
            )
        ] + [
            ConnPosArt(
                # x, y reused below
                x=cpo.x + (cpo.art.distance - 1 if not cpo.art.vertical else 0),
                y=cpo.y + (cpo.art.distance - 1 if cpo.art.vertical else 0),
                centered=cpo.centered,
                art=Line(
                    vertical=cpo.art.vertical,
                    distance=0,
                    connector="2",
                ),
            )
            for cpo in lines
            if not self.is_point_in_arts(
                # x, y reused above
                x=cpo.x + (cpo.art.distance - 1 if not cpo.art.vertical else 0),
                y=cpo.y + (cpo.art.distance - 1 if cpo.art.vertical else 0),
            )
        ]
        for arti in self.conncanvas.arts:
            assert isinstance(arti, ConnPosArt)
        arts_cc = [
            arti for arti in self.conncanvas.arts if isinstance(arti, ConnPosArt)
        ]
        self.conncanvas = ConnCanvas(arts=lines + arts_cc + line_ends)

    def remove_rs_text(self):
        """Remove the text of ProcessMini circles."""
        arts_new = [
            ConnPosArt(
                x=cpo.x,
                y=cpo.y,
                centered=cpo.centered,
                art=ProcessMini(text="")
                if isinstance(cpo.art, ProcessMini)
                else cpo.art,
            )
            for cpo in self.conncanvas.arts
        ]
        self.conncanvas = ConnCanvas(arts=arts_new)

    def get_name_input_other(
        self,
        input_other,
        allowed=None,
    ):
        """Get the name we should use for input_other."""
        if "detrendillum" in input_other.name_eso().lower():
            return text_from_dataitem(DetrendBase)
        elif (
            "ref_photom" in input_other.name_eso().lower()
            and input_other.name_eso() in self.thearts
        ):
            # TODO: Fix PHOTOMREF hack, e.g. check vertical dist. of dependency.
            return input_other.name_eso() + "2"
        else:
            return text_from_dataitem(input_other, allowed=allowed)

    def hack_around_overlapping_steps(self, step):
        """Sometimes RecipeSteps (DRL functions) overlap.

        The overlapping steps are moved explicitly.
        """
        # TODO: Find better way to deal with these otherwise overlapping steps.
        if step in [ApplyPersistence, ApplyNonlinearity]:
            if self.big:
                self.y_step += 6
                if step == ApplyNonlinearity:
                    self.y_step -= 18
            else:
                self.y_step -= 6
                if step == ApplyNonlinearity:
                    self.y_step -= 16

    def get_left_of_step_with_double_input(
        self,
        step,
        name_ycf_step,
    ):
        """Some steps have two inputs, e.g. CorrectGeomDistortion

        CorrectGeomDistortion has DistortionELT and DistortionWAM.
        """
        input_others = step.get_other_inputdata_full(
            never_include_previous=not self.big, include_raw=False
        )

        len_max = max(len(io.name_eso()) for io in input_others)
        # noinspection PyArgumentList
        posarts2 = [
            PosArt(
                x=0,
                y=i * 4,
                # TODO: Refactor all of this so it is not necessary to overrule
                #       the ditaaclass
                # art=get_ditaaclass(input_other, final=classes_final)(
                art=DataExternal(text=input_other.name_eso().ljust(len_max)),
            )
            for i, input_other in enumerate(input_others)
        ]
        ci2 = Canvas(arts=posarts2)
        # TODO: combine with the below
        assert name_ycf_step not in self.contained_ints_dict
        # Move down half the height of the input
        self.y_step += len(input_others) * 2
        self.contained_ints_dict[name_ycf_step] = ContainedInt(self.y_step)
        cpo_oi = ConnPosArt(
            x=self.contained_ints_dict["xpos_0"],
            y=self.contained_ints_dict[name_ycf_step],
            centered=True,
            art=ci2,
        )
        self.thearts[input_others[-1].name_eso()] = cpo_oi
        left = NeighbourArt(art=cpo_oi)
        # Move down other half the height of the input
        self.y_step += len(input_others) * 2
        return left

    def get_existing_names_for_left_of_step(
        self,
        input_other,
        name_input_other,
    ):
        """Check whether the left of step is already in the graph."""
        # Check whether the input is already in the graph.
        # Also check derived classes, especially if input is a Base clase,
        # e.g. Calbrated.photom is a PhotomBase, but preferably refer
        # to PhotomStdfield in the matrix because that makes it more complete.
        # TODO: However, this means that PHOTOMSTDFIELD ends up in the graph
        # even though it might be better to have PHOTOM there.
        derived_input_classes1 = input_other.get_derived_classes()
        derived_input_classes = [
            cls for cls in derived_input_classes1 if cls in self.classes_all
        ]

        # Some sanity checks to ensure such cases can only be introduced
        # deliberately. Expand checks when necessary.
        if derived_input_classes:
            assert input_other in [
                PhotomBase,
                DetrendFinalBase,
                StarcenterBase,
                DetrendBkgfield,
                DetrendFinalHCIBase,
                DistortionRefBase,
                DetrendFinalHCIScienceBase,
            ]
        else:
            assert input_other != PhotomBase

        # Also check base classes, e.g. DetrendWaffle might not be in the graph
        # but DetrendHCI (its superclass) might.
        base_input_classes = list(get_bases(input_other))
        names_to_check = [name_input_other] + [
            self.get_name_input_other(c)
            for c in derived_input_classes + base_input_classes
        ]
        # TODO: This is too similar to the logic in get_connposart_dataitem(),
        #       where raw_existing_names is created, so these need to be
        #       refactored.

        # E.g. prevent finding MASTERDARK_IMGDARK_RAW which is just
        # a repeat at the top of the big matrices.
        names_ok = [name for name in names_to_check if name in self.thearts]

        if not names_ok:
            derived_input_classes_to_check = derived_input_classes or [input_other]
            raw_labels = [r.name_eso() for r in derived_input_classes_to_check]
            names_ok1 = [
                name
                for name, posart in self.thearts.items()
                if isinstance(posart, ConnPosArt)
                and set(raw_labels) - set(posart.art.text.split()) == set()
            ]
            # TODO: Fix this hack to prevent finding MASTERDARK_IMGDARK_RAW
            names_ok = [
                name
                for name in names_ok1
                # This IMG/HDR + RAW in name check works.
                # This REF not in name is for backwards compatibility and
                # can perhaps better be removed.
                if not (("IMG" in name or "HDR" in name) and "RAW" in name)
                and "REF" not in name
            ]

        assert len(set(names_ok)) <= 1 or input_other in [
            StarcenterBase,
        ]
        if len(set(names_ok)) > 1 and names_ok[0] != name_input_other:
            # There are multiple to chose from, but not the name specified
            # so refuse to chose at all.
            names_ok = []

        return names_ok

    def get_left_of_step_normal(
        self,
        input_others,
        name_ycf_step,
    ):
        """The default way of getting the input to a RecipeStep."""

        # TODO: Fix DetrendFinal
        # assert len(input_others) == 1
        input_other = input_others[-1]
        name_input_other = self.get_name_input_other(
            input_other=input_other,
            allowed=self.classes_all,
        )

        # Check whether this input already exists in the graph.
        names_ok = self.get_existing_names_for_left_of_step(
            input_other=input_other,
            name_input_other=name_input_other,
        )

        oi: ConnPosArt = self.thearts.get(names_ok[0]) if names_ok else None

        if oi is not None:
            left = NeighbourArt(art=oi, distance=2)
        else:
            # Since this DataItem is not yet in the graph, it must have been
            # created elsewhere. So it is either Static, not created in any
            # pipeline. Those are hardcoded. Or it is created in another
            # pipeline, and get_ditaaclass has no way of knowing.
            clsn_oi = get_ditaaclass(input_other, final=self.classes_final)
            if clsn_oi != DataStatic:
                clsn_oi = DataExternal
            assert name_ycf_step not in self.contained_ints_dict
            self.contained_ints_dict[name_ycf_step] = ContainedInt(self.y_step)
            # noinspection PyArgumentList
            cpo_oi: ConnPosArt = ConnPosArt(
                x=self.contained_ints_dict["xpos_0"],
                y=self.contained_ints_dict[name_ycf_step],
                centered=True,
                art=clsn_oi(text=name_input_other),
            )
            self.thearts[name_input_other] = cpo_oi
            left = NeighbourArt(art=cpo_oi)

        return left

    def get_left_of_step(
        self,
        step,
        step_old,
        name_ycf_step,
    ):
        """Get what is left of this step."""

        self.hack_around_overlapping_steps(step)

        if step_old is not None:
            left = NeighbourArt(art=step_old)
        elif step in {
            CorrectGeomDistortion,
            ComputePhotomInScienceFilter,
            DeriveDpsicsDetrend,
        }:
            left = self.get_left_of_step_with_double_input(
                step=step,
                name_ycf_step=name_ycf_step,
            )
        elif input_others := step.get_other_inputdata_full(
            never_include_previous=not self.big, include_raw=False
        ):
            left = self.get_left_of_step_normal(
                input_others=input_others,
                name_ycf_step=name_ycf_step,
            )
        else:
            left = None

        # TODO: Fix this hack where we unset the line to the left.
        from micado.recipes.CreateVarDqLayers import CreateVarDqLayers

        if self.big and (left is not None) and (step == CreateVarDqLayers):
            left.line = ""

        return left

    def add_step(
        self,
        step,
        cf_xpos,
        cpo_prev,
    ):
        """Add a processing step to the matrix."""
        if step.name_eso() in self.thearts:
            step_old = self.thearts[step.name_eso()][-1]
        else:
            self.thearts[step.name_eso()] = []
            step_old = None

        name_ycf_step = "ypos_{}".format(step.name_eso())

        left = self.get_left_of_step(
            step=step,
            step_old=step_old,
            name_ycf_step=name_ycf_step,
        )
        if left is None and self.big:
            # A big matrix with all DRL Functions is being created.
            assert name_ycf_step not in self.contained_ints_dict
            self.contained_ints_dict[name_ycf_step] = ContainedInt(self.y_step)
            cf_ypos_step = self.contained_ints_dict[name_ycf_step]
        elif left is None:
            # A small matrix without DRL Functions is created, so only
            # include those that actually have input.
            if step.name_eso() in self.thearts and not len(
                self.thearts[step.name_eso()]
            ):
                # Remove DRL Function.
                self.thearts.pop(step.name_eso())
            # Return old data. Note that self.y_step is increased.
            self.y_step -= 3
            return None, cf_xpos, cpo_prev
        else:
            assert isinstance(left.art.y, ContainedInt)
            cf_ypos_step: ContainedInt = left.art.y
            self.y_step = max(cf_ypos_step.actual_int, self.y_step)

        assert isinstance(cf_ypos_step, ContainedInt)

        # TODO: Generalize this.
        # TODO: fix ":", it causes problems with the horizontal lines, so
        #       None is used for now.
        from micado.recipes.CreateVarDqLayers import CreateVarDqLayers

        line_style = (
            None
            if isinstance(cpo_prev.art, RecipeShort)
            and step != CreateVarDqLayers
            and cpo_prev.art.text != "micado_det_persistence"
            and "qacits" not in step.name_eso()
            else "|"
        )
        up = NeighbourArt(
            art=cpo_prev,
            distance=2
            if step.name_eso().endswith("_layers")
            or (cf_ypos_step - cpo_prev.bottom < 8)
            else None,
            line=line_style,
        )
        art = (
            Process(text=step.name_eso())
            if self.big
            else ProcessMini(text=step.name_eso())
        )
        cpo_step = ConnPosArt(
            x=cf_xpos,
            y=cf_ypos_step,
            art=art,
            centered=True,
            up=up,
            left=left,
        )
        self.thearts[step.name_eso()].append(cpo_step)
        cpo_prev_new = cpo_step
        self.y_step += 9 if self.big else 7
        return step, cf_xpos, cpo_prev_new

    def get_connposart_dataitem(
        self,
        do,
        cf_xpos,
        cpo_recipe_old,
        allowed_templates=None,
    ):
        """Get ConnPosArt for DataItem do."""
        assert hasattr(do, "recipe")
        assert hasattr(do, "get_raw_class")

        y_offset_steps_cheat = 0
        # TODO: The autoincrement of of ystep is broken, so all dataitems should
        #  be listed here...
        y_offset_steps_cheat_dict: Dict[Any, int] = {
            MasterDark: 35,
            MasterFlat: 54,
            IllumCorr: 70,
            Background: 85,
            Nonlinearity: 110,
            DetrendFinal: 120,
            DetrendFinalBase: 120,
            DistortionBase: 150,
            Astrom: 185,
            PhotomStdfield: 220,
            Calibrated: 240,
            CalibratedCatalog: 270,
            Regrid: 270,
            Stack: 270,
            # Starcenter is aligned such that the lines don't cross the boxes.
            StarcenterWaffle: 195,
            StarcenterPupil: 215,
            Staroffset: 180,
            CalibHCI: 240,
            CubeHCI: 270,
        }

        recipe = do.recipe.prop_type
        ditaa_recipe = RecipeShort(text=recipe.name_eso())

        step = recipe.process.prop_type
        steps = []
        while step:
            steps.append(step)
            step = step.get_previous()
            if not issubclass(step, RecipeStepBase):
                step = False

        raw = do.get_raw_class()
        progenitor = do.get_progenitor_class()

        template = raw.template.prop_type

        # Ensure that e.g. raw darks cannot be detrended.
        # raws = raw.get_derived_classes()
        raws = get_raws_derived(do)

        if do.__name__.endswith("Base"):
            # E.g. McdImgDetrendFinal could have any DetrendBase as input, but
            # it only makes sense for DetrendStdfield or Detrend.
            assert do in [
                DistortionBase,
                DetrendBase,
                DetrendFinalBase,
                DetrendHCIBase,
            ], do
            derived_dos = do.get_derived_classes()
            raws_derived = [di.get_progenitor_class() for di in derived_dos]
            raws_allowed = set(self.classes_all) & set(raws_derived)
        else:
            raws_allowed = set(self.classes_all)

        templates_raws = [template] + [ra.template.prop_type for ra in raws]
        allowed_templates_raw = set(templates_raws) & set(allowed_templates)

        text_template = text_from_template(template, allowed=allowed_templates_raw)
        ditaa_template = TemplateDitaa(text_template)

        # ditaa_raw = DataRaw(text=text_from_dataitem(raw))
        text_progenitor = text_from_dataitem(
            progenitor, allowed=raws_allowed, strict=True
        )
        clsn_prog = get_ditaaclass(progenitor, final=self.classes_final)
        ditaa_raw = clsn_prog(text=text_progenitor)

        # Try to find an existing art that corresponds to this progenitor.
        # That is, if DEPERSISTEDILLUM_IMG, DEPERSISTEDSCIENCE_IMG or
        # DEPERSISTEDSTDFIELD_IMG can be used as input, and there is an output
        # box with DEPERSISTEDDARK_IMG, DEPERSISTEDFLAT_IMG,
        # DEPERSISTEDILLUM_IMG, DEPERSISTEDSCIENCE_IMG, and
        # DEPERSISTEDSTDFIELD_IMG, then use that one.
        raw_labels = ditaa_raw.text.split()
        raw_existing_names = [
            name
            for name, posart in self.thearts.items()
            if isinstance(posart, ConnPosArt)
            and set(raw_labels) - set(posart.art.text.split()) == set()
        ]
        y = self.contained_ints_dict["ypos_raw" if self.big else "ypos_recipe"]
        if raw_existing_names and not self.big:
            # if raw_existing_names:
            # A sanity check. There can be more than 1 existing box that can
            # be reused, but let's make that explicit.
            # TODO: Perhaps remove this check? It seems to only be annoying
            #       and does not seem to actually help anything.
            assert len(raw_existing_names) == 1 or raw in [
                DetrendFinal,
                Opt,
                Sky,
                Science,
                CoroOffset,
                # Why is Stdfield there?
                Stdfield,
                SkySiaiBase,
            ]
            raw_existing_name = raw_existing_names[0]
            raw_existing = self.thearts[raw_existing_name]
            y = raw_existing.y
        else:
            raw_existing = None

        line = ":" if DASHED_HORIZONTAL else "|"
        cpo_raw = ConnPosArt(
            x=cf_xpos,
            y=y,
            art=ditaa_raw,
            centered=True,
            left=NeighbourArt(
                art=raw_existing,
                distance=2,
                line=line,
            )
            if raw_existing
            else None,
        )
        self.thearts[do.name_eso() + raw.name_eso()] = cpo_raw
        if self.big:
            cpo_template = ConnPosArt(
                x=cf_xpos,
                y=self.contained_ints_dict["ypos_template"],
                art=ditaa_template,
                centered=True,
            )
            self.thearts[f"{template.name_eso()}_{cf_xpos}"] = cpo_template
            cpo_raw.up = NeighbourArt(art=cpo_template, distance=2)

        y_offset_steps_cheat = y_offset_steps_cheat_dict.get(do, y_offset_steps_cheat)

        cpo_recipe = ConnPosArt(
            x=cf_xpos,
            y=self.contained_ints_dict["ypos_recipe" if self.big else "ypos_raw"],
            art=ditaa_recipe,
            centered=True,
            left=NeighbourArt(art=cpo_recipe_old, distance=2, line=None)
            if cpo_recipe_old is not None
            else None,
            up=NeighbourArt(art=cpo_raw, distance=2) if self.big else None,
        )
        if self.big:
            cpo_raw.down = NeighbourArt(art=cpo_recipe, distance=2)
        assert recipe.name_eso() not in self.thearts
        self.thearts[recipe.name_eso()] = cpo_recipe

        cpo_prev = cpo_recipe if self.big else cpo_raw
        self.y_step = (
            self.contained_ints_dict["ypos_recipe"] + 10 + y_offset_steps_cheat
        )
        cpos_steps = []
        for step in steps[::-1]:
            step, cf_xpos, cpo_prev = self.add_step(
                step,
                cf_xpos,
                cpo_prev,
            )
            if step is None:
                # big is False and there is no other input to this step
                continue
            cpos_steps.append(cpo_prev)
        ditaas_steps = [cpo.art for cpo in cpos_steps]

        clsn = get_ditaaclass(do, final=self.classes_final)
        # noinspection PyArgumentList
        ditaa_do = clsn(
            text=text_from_dataitem(do, allowed=self.classes_all, strict=True)
        )
        name_ycf_do = "ypos_{}".format(do.name_eso())
        assert name_ycf_do not in self.contained_ints_dict
        self.contained_ints_dict[name_ycf_do] = ContainedInt(self.y_step + 5)
        cpo_do = ConnPosArt(
            x=cf_xpos,
            y=self.contained_ints_dict[name_ycf_do],
            centered=True,
            art=ditaa_do,
            up=NeighbourArt(art=cpo_prev, distance=2),
        )
        widths = (
            [ditaa_do.width, ditaa_recipe.width]
            + [ditaa_raw.width]
            + [ditaa_template.width]
            + [d.width for d in ditaas_steps]
        )
        width_max = max(widths)
        return cpo_do, cpo_recipe, width_max

    def build_association_matrix(self):
        """Create Association Matrix."""
        allowed_templates = get_applicable_templates_from_dataitems(self.classes_all)
        cpo_recipe_old = None

        xpos_do = 30
        do: DataItem
        for doi, do in enumerate(self.classes_to_draw):
            assert hasattr(do, "recipe")
            assert hasattr(do, "get_raw_class")
            cf_xpos = ContainedInt(xpos_do)
            self.contained_ints_dict[f"xpos_{doi + 1}"] = cf_xpos
            # noinspection PyTypeChecker
            cpo_do, cpo_recipe, width_max = self.get_connposart_dataitem(
                do=do,
                cf_xpos=cf_xpos,
                cpo_recipe_old=cpo_recipe_old,
                allowed_templates=allowed_templates,
            )
            width_left = width_max - width_max // 2
            width_right = width_max // 2 + 1
            xpos_do += width_left
            xpos_do += width_right
            # One extra in order to be able to add arrows to the figures.
            xpos_do += 1
            cf_xpos += width_left
            assert do.name_eso() not in self.thearts
            self.thearts[do.name_eso()] = cpo_do

            cpo_recipe_old = cpo_recipe

        artsll = [a if isinstance(a, list) else [a] for a in self.thearts.values()]
        artsl = [a for al in artsll for a in al]
        self.conncanvas = ConnCanvas(arts=list(artsl))

    def generate_association_matrix(self):
        """Shuffle and add legend etc."""
        self.build_association_matrix()
        self.get_duosteps()

        legend_big = r"""

        +---------------+
        |               |
        | xxxxxxxxx_RAW |  Raw data
        |c999           |
        +---------------+

        +---------------+
        |               |
        | xx_EXT xx_CAT |  External data
        |c6CE           |
        +---------------+

        +---------------+
        |               |
        | xxxxxxxxx_IMG |  Intermediate data / Recipe data product
        |c47A           |
        +---------------+

        +---------------+
        |               |
        | xxxxxxxxx_IMG |  Final data
        |c283           |
        +---------------+



        +--------------+
        |              |
        | mcd_xxx_xxxx |  CPL Recipe
        |cB56          |
        +--------------+

        /-----\
        |cB56 |       DRL Function
        \-----/

        /-----\
        |cB56 +-- step to the right is identical
        \-----/

        /-----\
        |cB56 | - step to the right is different
        \-----/

    +--------------+
    |              |
    | micado_xx_xx |  CPL Recipe
    |cB56          |
    +--------------+



          |
        -----   data product on the left is not input to the recipe step below
          |

          |
        --+--   data product on the left is input to the recipe step below
          |

        """
        legend_small = r"""

        +---------------+
        |               |
        | xxxxxxxxx_RAW |  Raw data
        |c999           |
        +---------------+

        +---------------+
        |               |
        | xx_EXT xx_CAT |  External data
        |c6CE           |
        +---------------+

        +---------------+
        |               |
        | xxxxxxxxx_IMG |  Intermediate data / Recipe data product
        |c47A           |
        +---------------+

        +---------------+
        |               |
        | xxxxxxxxx_IMG |  Final data
        |c283           |
        +---------------+



        +--------------+
        |              |
        | mcd_xxx_xxxx |  CPL Recipe
        |cB56          |
        +--------------+



          |
        -----   data product on the left is not input to the recipe step below
          |

          |
        --+--   data product on the left is input to the recipe step below
          |

        """

        # TODO: Add these back once supported again.
        # ------ direct data dependency
        #
        # =----- indirect data dependency;
        #        actual input data is output from other recipe

        # This way it is only necessary to call .lines only once.
        lines = self.conncanvas.lines
        while lines := self.remove_line(lines):
            # Nothing to do here, because the line is already removed.
            pass

        # TODO: Perhaps these functions should not update self.conncanvas
        #       because they might make it inconsistent with self.thearts.
        self.add_lines()
        self.remove_rs_text()
        sss = str(self.conncanvas)

        # Remove empty columns
        cols = zip(*sss.splitlines())
        cols_crop = [col for col in cols if any(c != " " for c in col)]
        sss_crop = "".join("".join(c) + "\n" for c in zip(*cols_crop))

        # TODO: Fix hack to get dashed lines
        if DASHED_HORIZONTAL:
            sss_crop = sss_crop.replace("|-", "|=")
        else:
            sss_crop = sss_crop.replace("|-", "+-")

        if len(sss_crop.splitlines()) < 82:
            return sss_crop

        # TODO: Add legend back.
        s_legend = ""
        if self.legend:
            s_legend = legend_big if self.big else legend_small

        lineslegend = reversed(
            [
                lleg + lart[len(lleg) :] if lleg is not None else lart
                for lart, lleg in itertools.zip_longest(
                    reversed(sss_crop.splitlines()),
                    reversed(s_legend.splitlines()),
                )
            ]
        )
        return "\n".join(lineslegend)


def main():
    """Create matrix and write to file.

    Convert with

    java -jar ditaa-0.11.0-standalone.jar associationmatrix.ditaa -E \
        --no-shadows associationmatrix.png

    or

    java -jar ditaa-0.11.0-standalone.jar associationmatrix.ditaa -E \
        --no-shadows --svg associationmatrix.svg
    """
    name_classes = {
        # TODO: Fix Online HCI DPSICS or remove?
        # "qacits": (CLASSES_QACITS, [], CLASSES_QACITS_ALL),
        "hci": (CLASSES_HIGH_CONTRAST_IMAGING, [], CLASSES_HIGH_CONTRAST_IMAGING_ALL),
        "focal": (
            CLASSES_HIGH_CONTRAST_IMAGING_FOCAL,
            [],
            CLASSES_HIGH_CONTRAST_IMAGING_FOCAL_ALL,
        ),
        "pupil": (
            CLASSES_HIGH_CONTRAST_IMAGING_PUPIL,
            [],
            CLASSES_HIGH_CONTRAST_IMAGING_PUPIL_ALL,
        ),
        "si": (
            CLASSES_STANDARD_IMAGING,
            CLASSES_STANDARD_IMAGING_FINAL,
            CLASSES_STANDARD_IMAGING_ALL,
        ),
        "ai": (
            CLASSES_ASTROMETRIC_IMAGING,
            CLASSES_ASTROMETRIC_IMAGING_FINAL,
            CLASSES_ASTROMETRIC_IMAGING_ALL,
        ),
        "sismall": (
            CLASSES_STANDARD_IMAGING_SMALL,
            CLASSES_STANDARD_IMAGING_FINAL,
            CLASSES_STANDARD_IMAGING_ALL,
        ),
        "hcismall": (
            CLASSES_HIGH_CONTRAST_IMAGING_SMALL,
            [],
            CLASSES_HIGH_CONTRAST_IMAGING_ALL,
        ),
        "aismall": (
            CLASSES_ASTROMETRIC_IMAGING_SMALL,
            CLASSES_ASTROMETRIC_IMAGING_FINAL,
            CLASSES_ASTROMETRIC_IMAGING_ALL,
        ),
        "reduced": (
            CLASSES_UPTO_REDUCED,
            CLASSES_UPTO_REDUCED_FINAL,
            CLASSES_UPTO_REDUCED_ALL,
        ),
    }
    for name, (classes, classes_final, classes_all) in name_classes.items():
        for big in [
            False,
            True,
        ]:
            association = AssociationMatrix(
                classes_to_draw=classes,
                classes_final=classes_final,
                classes_all=classes_all,
                big=big,
            )
            s_association = association.generate_association_matrix()
            print(s_association)
            postfix_name = "_" + name if name != "si" else ""
            postfix_big = "_big" if big else ""
            path_associationmatrix = (
                PATH_DOCS_GENERATED
                / f"associationmatrix{postfix_name}{postfix_big}.ditaa"
            )
            with open(path_associationmatrix, "w") as f:
                f.write(s_association)


if __name__ == "__main__":
    main()
