"""Script generates an overview of the node it is running on"""

import logging
import os
import socket
import subprocess
import sys
import time
import unittest
import zlib

from collections import defaultdict
from datetime import datetime
from glob import glob
from optparse import OptionParser
from pprint import pprint, pformat

from common.config.Environment import Env
from common.util.OptionsParser import OptionsParser
from common.util.Processor import ProcessorError

import micado

from micado.util.Task import Task


class NodeInfoJob(Task):

    # expected processing time
    dpu_time = 60  # seconds

    parser = OptionsParser(__name__)
    parser.add_option("s", "sleep", 0, "The sleep time, running after the job")
    parser.add_option("m", "multiple", 0, "Run the job multiple times")

    def execute(self):
        """run job"""
        self.show()
        if self.sleep > 0:
            time.sleep(self.sleep)

    def get_process_time(self):
        """return the default processing time"""
        return self.dpu_time + self.sleep  # [seconds]

    def get_result(self):
        return [1]

    def get_process_result(self):
        return [2]

    def show(self):
        """show the node info"""

        logging.getLogger("node_overview").info("---=== Node info ===---")
        logging.getLogger("node_overview").info("sys.executable %s" % sys.executable)
        logging.getLogger("node_overview").info("fqdn %s" % socket.getfqdn())
        logging.getLogger("node_overview").info("dpu %s" % Env["dpu_name"])

        result = defaultdict(list)

        for dirname, dirnames, filenames in os.walk(micado.__path__[0]):
            for filename in filenames:
                if filename.endswith(".py"):
                    filepath = os.path.join(dirname, filename)
                    result[dirname].append(
                        (
                            datetime.fromtimestamp(os.path.getmtime(filepath)).strftime(
                                "%Y-%m-%d %H:%M:%S"
                            ),
                            filepath,
                        )
                    )

        logging.getLogger("node_overview").info("---=== git log files ===---")
        path = os.path.join(micado.__path__[0], "..")
        for filename in glob(os.path.join(path, ".*.log")):
            filepath = os.path.join(path, filename)
            logging.getLogger("node_overview").info(filepath)
            logging.getLogger("node_overview").info(open(filepath).read())

        logging.getLogger("node_overview").info(
            "---=== files and dates per directory ===---"
        )
        for dirname in sorted(result):
            logging.getLogger("node_overview").info("%s" % dirname)
            for date, filepath in sorted(result[dirname]):
                logging.getLogger("node_overview").info("%s  %s" % (date, filepath))

        logging.getLogger("node_overview").info("---=== os.environ ===---")
        for name, value in sorted(os.environ.items()):
            if "password" in name:
                value = "********"
            logging.getLogger("node_overview").info("  %s = %s" % (name, value))

        logging.getLogger("node_overview").info("---=== Env ===---")
        for name, value in sorted(Env.items()):
            if "password" in name:
                value = "********"
            logging.getLogger("node_overview").info("  %s = %s" % (name, value))

        return 0


def handle_option_multiple(options, dpu, job_ids):
    """handle multiple jobs"""
    from micado.util.Pipeline import NodeInfo

    if options.parallel:
        for count in range(options.multiple):
            job_ids.append(dpu.run("nodeinfo", **args))
        maxwait = 1000
    else:
        if options.multiple > 2:
            NodeInfo.task_sequences["nodeinfos"] += (options.multiple - 2) * [
                "nodeinfos"
            ]
        job_ids.append(dpu.run("nodeinfos", **args))
        maxwait = 10000
    return maxwait


def run(options, args):
    """submit NodeInfoJob to node and print log"""
    from micado.util.Processor import Processor

    dpu = Processor(Env["dpu_name"], request_logs=False)

    job_ids = []
    if options.multiple > 1:
        maxwait = handle_option_multiple(options, dpu, job_ids)
    else:
        job_ids.append(dpu.run("nodeinfo", **args))
        maxwait = 600
        if "sleep" in args:
            maxwait += args["sleep"]
    for job_id in job_ids:
        if job_id is None:
            logging.getLogger("node_overview").error(
                "DPU %s did not return a job id" % Env["dpu_name"]
            )
        else:
            try:
                dpu.wait_for_jobs(job_id, maxwait=maxwait)
                logging.getLogger("node_overview").info("getsyslogs({})".format(job_id))
                syslogs = dpu.dpu_client.getsyslogs(job_id)
                if syslogs:
                    for syslog in syslogs[0]:
                        logging.getLogger("node_overview").info("".join(syslog))
                for job_dict in dpu.dpu_client.retrievejobs(job_id):
                    for log, job in zip(job_dict["DPU_LOGS"], job_dict["DPU_JOBS"]):
                        logging.getLogger("node_overview").info("Job log:")
                        logging.getLogger("node_overview").info(log)
            except ProcessorError as e:
                logging.getLogger("node_overview").error("ProcessorError; %s" % e)
                dpu.cancel_job(job_id)
    return 0


if __name__ == "__main__":
    parser = OptionParser(usage="usage: awe %prog [options]")
    parser.add_option(
        "-u",
        "--dpu",
        default=False,
        action="store_true",
        help="Switch to run node info on dpu",
    )
    parser.add_option(
        "-n",
        "--dpu_debug_nocleanup",
        default=False,
        action="store_true",
        help="Switch to not do cleanup on dpu",
    )
    parser.add_option(
        "-l",
        "--sleep",
        default=0,
        type=int,
        help="Time in seconds to sleep after job runs using DPU",
    )
    parser.add_option(
        "-t",
        "--time",
        default=60,
        type=int,
        help="Estimated job time in seconds, excluding sleep",
    )
    parser.add_option(
        "-m", "--multiple", default=1, type=int, help="Run multiple NodeInfo jobs"
    )
    parser.add_option(
        "-p",
        "--parallel",
        default=False,
        action="store_true",
        help="Run multiple jobs in parallel",
    )
    (options, args) = parser.parse_args()

    # init logging to INFO
    logging.getLogger().addHandler(logging.StreamHandler(stream=sys.stdout))
    logging.getLogger().setLevel(logging.INFO)

    # explicit import needed, otherwise the class atributes are not updated
    from micado.toolbox.node_overview import NodeInfoJob

    if options.dpu_debug_nocleanup:
        NodeInfoJob.dpu_debug_nocleanup = True

    NodeInfoJob.dpu_time = options.time

    args = {}
    if options.sleep > 0:
        args["sleep"] = options.sleep
    args["multiple"] = options.multiple

    if options.dpu:
        sys.exit(run(options, args))
    else:
        if options.multiple > 1:
            print("Option mutiple ignored in local mode")
        sys.exit(NodeInfoJob(**args).execute())
    sys.exit(0)
