# -*- coding:utf-8 -*-
"""Create an ESO OCA file from the Python classes."""
import math
import operator
from collections import OrderedDict
from dataclasses import dataclass, field
from typing import Union, Type, Any, List, Tuple

from micado.dataitems.DetrendFinal import DetrendFinal
from micado.main.commonwrapper import link_list_property, persistent_property
from micado.dataitems.Bkgfield import Bkgfield

from micado.dataitems.CoroBase import CoroBase
from micado.dataitems.DetrendBkgfield import DetrendBkgfield
from micado.dataitems.DetrendFinalSiaiBase import DetrendFinalSiaiBase
from micado.dataitems.PhotomBase import PhotomBase
from micado.dataitems.PhotomScience import PhotomScience
from micado.dataitems.PhotomStdfield import PhotomStdfield
from micado.dataitems.DpsicsOptSciBase import DpsicsOptSciBase
from micado.main.Opt import Opt
from micado.main.RecipeStepExtra import name_eso_from_property
from micado.main.Sky import Sky
from micado.dataitems.Stack import Stack
from micado.main.Cat import Cat
from micado.main.DataItem import DataItem
from micado.main.Raw import Raw
from micado.main.Img import Img
from micado.main.SkyHCI import SkyHCI
from micado.main.SkyHCIOnline import SkyHCIOnline
from micado.main.SkySiaiBase import SkySiaiBase
from micado.main.Tab import Tab
from micado.main.Ext import Ext
from micado.toolbox.augment_backward_chaining import augment_backward_chaining

# aweimports fills the classcache.
# noinspection PyUnresolvedReferences
import micado.main.aweimports  # noqa NOSONAR

augment_backward_chaining()


class OCAClassification:
    """OCA Classification Rule."""

    def __new__(cls, dataitem):
        """Factory method to create OCAClassification of the right class."""
        if issubclass(dataitem, Raw):
            classification = object.__new__(OCAClassificationRaw)
        elif issubclass(dataitem, (Ext, Cat)):
            classification = object.__new__(OCAClassificationExt)
        elif issubclass(dataitem, (Img, Tab)):
            classification = object.__new__(OCAClassificationPro)
        else:
            raise ValueError(f"Cannot create OCA rule for {dataitem.__name__}")

        return classification


class OCAClassificationRaw(OCAClassification):
    """OCA Classification Rule for Raw data."""

    def __init__(self, dataitem: Type[DataItem]):
        """Constructor.

        Args:
            dataitem: DataItem to classify.
        """
        self.dataitem = dataitem
        # The base raw class itself does not trigger a recipe all by itself.
        self.target = issubclass(dataitem, Raw) and dataitem != Raw

    @property
    def where(self):
        """where-clause of this classification."""
        # NODEFAULT is the value used by ESO tsf files if there is no default.
        catg = self.dataitem.dpr_catg_c()
        if catg in {"NODEFAULT"}:
            catg = ""
        return f'''    DPR.CATG like "{catg}%"
    and DPR.TECH like "{self.dataitem.dpr_tech_c()}%"
    and DPR.TYPE like "{self.dataitem.dpr_type_c()}%"'''.replace(
            "%%", "%"
        )

    def __str__(self):
        """The OCA rule in text."""
        # TODO: Add "and INSTRUME like "MICADO%" or something like that?
        s_target = '    REFLEX.TARGET = "T";\n' if self.target else ""
        return f"""if
{self.where} then
{{
    DO.CATG = "{self.dataitem.name_oca()}";
    REFLEX.CATG = "{self.dataitem.name_oca()}";
{s_target}}}
"""

    @property
    def where_edps(self):
        """EDPS where clause."""
        # TODO: Use .like
        s_edps = f'f["dpr.catg"] == "{self.dataitem.dpr_catg_c()}" and f["dpr.tech"] == "{self.dataitem.dpr_tech_c()}" and f["dpr.type"] == "{self.dataitem.dpr_type_c()}"'
        return s_edps

    def str_edps(self):
        """The EDPS rule in Python."""
        s_edps = f"""
def is_{self.dataitem.name_oca().lower()}(f):
    return {self.where_edps}

{self.dataitem.name_oca().lower()}_class = ClassificationRule("{self.dataitem.name_eso()}", is_{self.dataitem.name_oca().lower()})

{self.dataitem.name_oca().lower()} = (
    data_source()
    .with_classification_rule({self.dataitem.name_oca().lower()}_class)
    .build()
)
"""
        return s_edps


class OCAClassificationExt(OCAClassification):
    """OCA Classification Rule for External data."""

    def __init__(self, dataitem):
        """Constructor.

        Args:
            dataitem: DataItem to classify.
        """
        self.dataitem = dataitem
        self.target = False

    @property
    def where(self):
        """where-clause of this classification."""
        return f'''    PRO.CATG like "{self.dataitem.name_oca()}%"'''.replace("%%", "%")

    def __str__(self):
        """The OCA rule in text."""
        # TODO: Add "and INSTRUME like "%MICADO" or something like that?
        return f"""if
{self.where} then
{{
    DO.CATG = "{self.dataitem.name_oca()}";
    REFLEX.CATG = "{self.dataitem.name_oca()}";
}}
"""

    @property
    def where_edps(self):
        """EDPS where clause."""
        # TODO: Use .like
        s_edps = f'f["pro.catg"] == "{self.dataitem.name_oca()}"'
        return s_edps

    def str_edps(self):
        """The EDPS rule in Python."""
        # TODO: Add with_match_function
        s_edps = f"""
def is_{self.dataitem.name_oca().lower()}(f):
    return {self.where_edps}

{self.dataitem.name_oca().lower()}_class = ClassificationRule("{self.dataitem.name_eso()}", is_{self.dataitem.name_oca().lower()})

{self.dataitem.name_oca().lower()} = (
    data_source()
    .with_classification_rule({self.dataitem.name_oca().lower()}_class)
    .build()
)
"""
        return s_edps


class OCAClassificationPro(OCAClassificationExt):
    """OCA Classification Rule for Processed data."""

    def str_edps(self):
        """The EDPS rule in Python."""
        return ""


@dataclass
class OCAOrganization:
    """OCA Organization Rule."""

    progenitor: Type[DataItem]
    dataitem: Type[DataItem]
    classification: OCAClassification = None
    groupby: List[Any] = field(default_factory=list)

    def __post_init__(self):
        """Classify the progenitor of this DataItem."""
        # Processed data as progenitor is handled by OCARules.
        self.classification = OCAClassification(self.progenitor)

    def __str__(self):
        """The OCA rule in text."""
        sgroupby = ", ".join(
            name_eso_from_property(el, dotted=True) for el in self.groupby
        )

        sgroupbyclause = f"group by {sgroupby}" if sgroupby else ""
        return f"""select execute(ACTION_{self.dataitem.name_oca()}) from inputFiles
    where DO.CATG like "{self.progenitor.name_oca()}%"
    {sgroupbyclause};
"""


class OCAAssociation:
    """OCA Association/Action Rule."""

    def __init__(self, dataitem: Type[DataItem]):
        """Constructor.

        Args:
            dataitem: DataItem to classify.
        """
        self.dataitem = dataitem
        self.recipe = self.dataitem.recipe.prop_type
        self.selectfiles = {
            do: OCASelectFile(
                cls=do,
                where=[
                    OCASelectFileWhere(
                        cls=do,
                        where=wherei,
                    )
                    for wherei in dataitem.lineage.get(prop, [])
                ],
                min_ret=0 if isinstance(prop, link_list_property) else 1,
                max_ret=999 if isinstance(prop, link_list_property) else 1,
            )
            for (prop, do) in dataitem.get_inputdata_full().items()
            if str((prop, do)) != str(dataitem.get_progenitor_attr_class())
            # Okay. The str()'s above are necessary, because otherwise two
            # properties will be compared, and that is not possible if a
            # database backend is set. E.g.
            # >>> Dark.template == Dark.template
            # RuntimeError: Select object cannot be used without an
            #    operator in conditional statements
        }
        prop, do = dataitem.get_progenitor_attr_class()
        groupby_clauses = dataitem.lineage.get(prop, [])
        # Some defensive programming.
        assert groupby_clauses
        for clause in groupby_clauses:
            assert all(
                isinstance(el, persistent_property) or el is operator.__eq__
                for el in clause
            )
        groupbys = [
            a
            for clausea in groupby_clauses
            for a in [
                el
                for el in clausea
                if isinstance(el, persistent_property)
                and el.cls == self.dataitem.get_progenitor_class()
            ]
        ]
        self.organization = OCAOrganization(
            progenitor=self.dataitem.get_progenitor_class(),
            dataitem=self.dataitem,
            groupby=groupbys,
        )

    def __str__(self):
        """The OCA rule in text."""
        s_selectfiles = "\n\n".join(str(sf) for sf in self.selectfiles.values())
        return f"""action ACTION_{self.dataitem.name_oca()}
{{
{s_selectfiles}

    recipe {self.recipe.name_eso()};

    product {self.dataitem.name_oca()} {{
        PRO.CATG = "{self.dataitem.name_oca()}";
        DO.CATG = "{self.dataitem.name_oca()}";
        REFLEX.CATG = "{self.dataitem.name_oca()}";
    }}
}}
"""

    def str_edps(self):
        """EDPS rule in Python."""
        s_withs = "\n".join(sf.str_edps() for sf in self.selectfiles.values())
        theprog = self.organization.progenitor
        # TODO: Add support for Base classes for EDPS.
        if theprog == DetrendFinalSiaiBase:
            theprog = DetrendFinal
        return f"""
{self.dataitem.name_oca().lower()} = (
    task('{self.recipe.name_eso()}')
    .with_main_input({theprog.name_oca().lower()})
{s_withs}
    .build()
)
"""


@dataclass
class OCASelectFileWhere:
    """A where clause for a SelectFile."""

    cls: Type[DataItem]
    where: Tuple[Any]

    def _str_from_element(self, el):
        str_from_operator = {
            operator.__eq__: "==",
            operator.__ne__: "!=",
            operator.__lt__: "<",
            operator.__le__: "<=",
            operator.__gt__: ">",
            operator.__ge__: ">=",
        }
        if el in str_from_operator:
            return str_from_operator[el]
        elif el.cls == self.cls:
            return f"inputFiles.{name_eso_from_property(el, dotted=True)}"
        else:
            return f"inputFile.{name_eso_from_property(el, dotted=True)}"

    def __str__(self):
        """And clause for selectfile."""
        elements = [self._str_from_element(el) for el in self.where]
        s_clause = " ".join(elements)
        assert "inputFiles." in s_clause
        assert "inputFile." in s_clause
        return s_clause


@dataclass
class OCASelectFile:
    """OCA select file clause."""

    cls: Type[DataItem]
    where: List[OCASelectFileWhere]
    # TODO: support different minRet and maxRet, it is in the vodsl.
    min_ret: int = 1
    max_ret: int = 1

    # TODO: add clauses

    def __str__(self):
        """The OCA rule in text."""
        # Selection is done from calibFiles, because this seems to contain both
        # already existing FITS files as well as FITS files generated earlier
        # in the pipeline. Most OCA rulesets only select from calibFiles in
        # their association rules (and only from inputFiles in their
        # organization rules.
        clauses = [f'DO.CATG like "{self.cls.name_oca()}%"'] + [
            str(wherei) for wherei in self.where
        ]
        joined_clauses = "\n        AND ".join(clauses)
        return f"""    minRet = {self.min_ret}; maxRet = {self.max_ret};
    select file as {self.cls.name_oca()} from calibFiles where
        {joined_clauses};"""

    def str_edps(self):
        """EDPS rules."""
        # TODO: Add support for base classes to EDPS rules.
        thecls = self.cls
        if thecls == PhotomBase:
            thecls = PhotomScience
        return f"""    .with_associated_input({thecls.name_oca().lower()})"""


class OCARules:
    """A full set of OCA rules."""

    def __init__(self, allowed=None):
        """Constructor."""
        self.classifications = {}
        self.organizations = {}
        self.associations = {}
        self.allowed = allowed

        if allowed:
            for dataitem in allowed:
                self.add_dataitem_classification(dataitem)

    def add_dataitem_pro(self, dataitem: Type[Union[Img, Tab]], depth=math.inf):
        """Add processed dataitem."""
        assert issubclass(dataitem, DataItem)
        if not depth:
            return
        if dataitem not in self.associations:
            self.add_dataitem(dataitem.get_progenitor_class(), depth=depth - 1)
            association = OCAAssociation(dataitem)
            for do in association.selectfiles:
                self.add_dataitem(do, depth=depth - 1)
            self.associations[dataitem] = association
            self.organizations[dataitem] = association.organization

    def add_dataitem_classification(self, dataitem: Type[DataItem]):
        """Add a DataItem classification."""
        if dataitem not in self.classifications:
            classification = OCAClassification(dataitem)
            # Only allow Raw to have a match-everything wildcard.
            if 'like "%"' in classification.where and dataitem != Raw:
                return
            if classification.where in [c.where for c in self.classifications.values()]:
                raise ValueError("Duplicate where clause")
            self.classifications[dataitem] = classification

    def add_dataitem(self, dataitem: Type[DataItem], depth=math.inf):
        """Add a dataitem and its dependencies to the OCA rules."""
        if self.allowed is not None:
            if dataitem not in self.allowed:
                return

        # Processed data should have its dependencies added.
        if issubclass(dataitem, (Img, Tab)):
            # Base classes should not be added themselves.
            if "Base" in dataitem.__name__:
                # Defensive programming, only PhotomBase is expected here
                # from Calibrated. And DetrendFinalSiaiBase.
                assert dataitem in [PhotomBase, DetrendFinalSiaiBase]
            else:
                self.add_dataitem_pro(dataitem, depth=depth)

        # Get derived classes.
        dataitems_derived = dataitem.get_derived_classes()
        # TODO: Add hacks here too to create background from multiple
        #  observations. Or fix hacks.
        if dataitem in [
            Bkgfield,
            DetrendBkgfield,
        ]:
            dataitems_derived = []

        if len(dataitems_derived) > 1:
            assert dataitem in [
                PhotomBase,
                Raw,
                Opt,
                Sky,
                SkySiaiBase,
                SkyHCI,
                CoroBase,
                DpsicsOptSciBase,
                SkyHCIOnline,
                # TODO: See whether this still makes sense.
                DetrendFinalSiaiBase,
            ] or dataitem.__name__.endswith("Base"), dataitem

        # Add the class itself.
        self.add_dataitem_classification(dataitem)

        if dataitem == PhotomBase:
            assert PhotomStdfield in dataitems_derived
            assert PhotomScience in dataitems_derived

        # Add the derived classes.
        for dataitem_der in dataitems_derived:
            self.add_dataitem(dataitem_der, depth=depth)

    def sort(self):
        """Sort the classification rules."""

        def key(do_c):
            """Count number of underscores and whether RAW is there."""
            do, _ = do_c
            return do.name_oca().count("_"), "RAW" not in do.name_oca()

        classification_list = sorted(
            self.classifications.items(),
            key=key,
        )
        self.classifications = OrderedDict(classification_list)

    def __str__(self):
        """The OCA rules in text. Content of .oca file."""
        self.sort()
        s_classifications = "\n".join(str(r) for r in self.classifications.values())
        s_organizations = "\n".join(str(r) for r in self.organizations.values())
        s_associations = "\n".join(str(r) for r in self.associations.values())
        return f"""// CLASSIFICATION

{s_classifications}

// ORGANIZATION

{s_organizations}

// ASSOCIATION

{s_associations}
"""

    def str_edps(self):
        """Return the EDPS string."""
        self.sort()
        s_classifications = "\n".join(
            r.str_edps() for r in self.classifications.values()
        )
        # TODO: Is it useful to add separate Organization rules?
        # s_organizations = "\n".join(r.str_edps() for r in self.organizations.values())
        # TODO: There are many duplicates now!
        s_associations = "\n".join(r.str_edps() for r in self.associations.values())
        return f"""
from edps.generator.task_builder import data_source, task
from edps.generator.task import ClassificationRule, ProductClassificationRule

{s_classifications}

{s_associations}
"""


def main():
    """Generate the main OCA Rules."""
    my_rules = OCARules()
    my_rules.add_dataitem(Stack)
    print(
        """// OCA rules for MICADO Imaging.
//
// This file is automatically """
        + """generated. Do not modify!
//
// Guiding principles for the Imaging OCA Rules:
// - The rules are automatically generated from the digital design.
// - DO.CATG, REFLEX.CATG and PRO.CATG have the same values where applicable.
// - All FITS files that are used or produced in the pipeline have a
//   classification rule that sets their 'DO.CATG', even when they are the
//   final output of the pipeline.
// - Every ACTION gives the PRO.CATG of the output that it produces. This
//   ensures that the full pipeline can be reconstructed from just the OCA
//   rules. Recipes with multiple possible outputs have multiple actions.
// - Classification rules are sufficiently precise that organisation and
//   association rules only have to select on DO.CATG.
// - Classification rules are hierarchical, so like-queries can be used to
//   find groups of related-but-different types of data products. (As a result
//   the DO.CATG values are quite long.)
// - Organization rules always select from inputFiles, because that is the only
//   data-set allowed there. Association rules always select from calibFiles,
//   because that includes the inputFiles as well. rawFiles are not queried.
"""
    )
    print(my_rules)


if __name__ == "__main__":
    main()
