"""
    Module with job definition for simulating raw data

    These are all the Raw classes, that have no derived classes:
         <class 'micado.dataitems.Dark.Dark'>: 1,
         <class 'micado.dataitems.Flat.Flat'>: 1,
         <class 'micado.dataitems.Nonlin.Nonlin'>: 1,
         <class 'micado.dataitems.Bkgfield.Bkgfield'>: 1,
         <class 'micado.dataitems.Illum.Illum'>: 1,
         <class 'micado.dataitems.CoroOffset.CoroOffset'>: 1,
         <class 'micado.dataitems.Astromfield.Astromfield'>: 1,
         <class 'micado.dataitems.Wam.Wam'>: 1,
         <class 'micado.dataitems.ScienceCoroWaffle.ScienceCoroWaffle'>: 1,
         <class 'micado.dataitems.ScienceCoroPupil.ScienceCoroPupil'>: 1,
         <class 'micado.dataitems.Science.Science'>: 1,
         <class 'micado.dataitems.ScienceAi.ScienceAi'>: 1,
         <class 'micado.dataitems.Stdfield.Stdfield'>: 1,
         <class 'micado.dataitems.Cam.Cam'>: 1,
         <class 'micado.dataitems.CoroBackground.CoroBackground'>: 1,
         <class 'micado.dataitems.DpsicsOpt.DpsicsOpt'>: 1

    These are all the Raw classes, that have 1 or more derived classes:
         <class 'micado.main.Raw.Raw'>: 25,
         <class 'micado.main.Opt.Opt'>: 23,
         <class 'micado.main.Sky.Sky'>: 17,
         <class 'micado.main.SkySiaiBase.SkySiaiBase'>: 3,
         <class 'micado.main.SkyHCI.SkyHCI'>: 9,
         <class 'micado.main.SkyHCIOnline.SkyHCIOnline'>: 3,
         <class 'micado.dataitems.CoroBase.CoroBase'>: 5,
         <class 'micado.dataitems.WamCamBase.WamCamBase'>: 3,
         <class 'micado.dataitems.DpsicsOptSciBase.DpsicsOptSciBase'>: 2,


    Start with these raw classes:

         <class 'micado.dataitems.Dark.Dark'>: 1,
         <class 'micado.dataitems.Flat.Flat'>: 1,
         <class 'micado.dataitems.Science.Science'>: 1,
         <class 'micado.dataitems.Stdfield.Stdfield'>: 1,
         <class 'micado.dataitems.Illum.Illum'>: 1,

    and optional:
         <class 'micado.dataitems.Bkgfield.Bkgfield'>: 1,
         <class 'micado.dataitems.Nonlin.Nonlin'>: 1,
"""

import copy
from datetime import datetime

from common.util.OptionsParser import OptionsParser
from micado.dataitems.Dark import Dark
from micado.dataitems.Flat import Flat
from micado.dataitems.Illum import Illum
from micado.dataitems.Science import Science
from micado.dataitems.Stdfield import Stdfield
from micado.hardware.DarkCurrent import default_darkcurrent
from micado.hardware.Detector import Detector
from micado.hardware.DetectorArray import DetectorArray, default_detectorarray
from micado.hardware.FieldConstantPSF import FieldConstantPSF
from micado.hardware.FilterWheel import FilterWheel
from micado.hardware.Instrument import Instrument
from micado.hardware.LinearityCurve import default_linearitycurve
from micado.hardware.NonCommonPathAberration import NonCommonPathAberration
from micado.hardware.PoorMansHxRGReadoutNoise import (
    default_poormanshxrgreadoutnoise,
)
from micado.hardware.QuantumEfficiencyCurve import (
    default_quantumefficiencycurve,
)
from micado.hardware.ReferencePixelBorder import (
    default_referencepixelborder,
)
from micado.hardware.RelayOptics import RelayOptics
from micado.hardware.ShotNoise import default_shotnoise
from micado.hardware.SummedExposure import default_summedexposure
from micado.hardware.SurfaceList import SurfaceList
from micado.main.ObservingBlock import ObservingBlock
from micado.main.Raw import Raw
from micado.templates.MicadoImgCalDark import MicadoImgCalDark
from micado.templates.MicadoImgCalFlat import MicadoImgCalFlat
from micado.templates.MicadoImgCalZeropoint import MicadoImgCalZeropoint
from micado.templates.MicadoImgObs import MicadoImgObs
from micado.templates.MicadoImgTecIllum import MicadoImgTecIllum

from .Task import Task, TaskError, deserialize_object

# TODO: Perhaps move these default test detectors and array somewhere generic.
# TODO: Or also generate it automatically.
mini_default_detector_array_list_list = [
    Detector(
        id=1,
        x_cen=-63.94,
        y_cen=63.94,
        x_size=32,
        y_size=32,
        pixel_size=0.015,
        angle=0.0,
        gain=1.0,
    ),
    Detector(
        id=2,
        x_cen=0.0,
        y_cen=63.94,
        x_size=32,
        y_size=32,
        pixel_size=0.015,
        angle=270.0,
        gain=1.0,
    ),
    Detector(
        id=3,
        x_cen=63.94,
        y_cen=63.94,
        x_size=32,
        y_size=32,
        pixel_size=0.015,
        angle=180.0,
        gain=1.0,
    ),
    Detector(
        id=6,
        x_cen=-63.94,
        y_cen=0.0,
        x_size=32,
        y_size=32,
        pixel_size=0.015,
        angle=0.0,
        gain=1.0,
    ),
    Detector(
        id=5,
        x_cen=0.0,
        y_cen=0.0,
        x_size=32,
        y_size=32,
        pixel_size=0.015,
        angle=180.0,
        gain=1.0,
    ),
    Detector(
        id=4,
        x_cen=67.94,
        y_cen=0.0,
        x_size=32,
        y_size=32,
        pixel_size=0.015,
        angle=180.0,
        gain=1.0,
    ),
    Detector(
        id=7,
        x_cen=-63.94,
        y_cen=-63.94,
        x_size=32,
        y_size=32,
        pixel_size=0.015,
        angle=0.0,
        gain=1.0,
    ),
    Detector(
        id=8,
        x_cen=0.0,
        y_cen=-63.94,
        x_size=32,
        y_size=32,
        pixel_size=0.015,
        angle=90.0,
        gain=1.0,
    ),
    Detector(
        id=9,
        x_cen=63.94,
        y_cen=-63.94,
        x_size=32,
        y_size=32,
        pixel_size=0.015,
        angle=180.0,
        gain=1.0,
    ),
]

mini_default_detectorarray = DetectorArray(
    name="",
    image_plane_id=0,
    temperature=-230.0,
    readout="CDS",
    detector_array_list=mini_default_detector_array_list_list,
    dark_current=default_darkcurrent,
    shot_noise=default_shotnoise,
    qe_curve=default_quantumefficiencycurve,
    linearity=default_linearitycurve,
    exposure_action=default_summedexposure,
    readout_noise=default_poormanshxrgreadoutnoise,
    reference_pixel_border=default_referencepixelborder,
)


class SimulateRawJobError(TaskError):
    """error raised in case of error"""

    pass


class SimulateRawJob(Task):
    """job runs simulate raw data"""

    # duration of job
    processing_time = 60  # seconds

    ## Defaults for these must be set in derived classes
    # the target to make
    target = Raw
    # default template class to use
    template_default = None
    # observing block target name
    observingblock_target = None

    raw_parser = OptionsParser(__name__)

    raw_parser.add_option(
        "c",
        "commit",
        False,
        "Commit switch, set True to commit results to database and data server",
    )
    raw_parser.add_option(
        "s",
        "skip_store",
        False,
        "Do not store on dataserver. Only to be used for testing.",
    )
    raw_parser.add_option(
        "utd",
        "use_test_detectors",
        False,
        "Use small test detectors. Only to be used for testing.",
    )
    raw_parser.add_option(
        "n",
        "nr_raws",
        5,
        "Number of raws exposures to simulate",
    )
    raw_parser.add_option(
        "t",
        "template",
        None,
        "The template; either give as object or as (class, object_id)",
    )
    raw_parser.add_option(
        "o",
        "observing_block",
        None,
        "The Observing Block; either give as object or as (class, object_id)",
    )
    raw_parser.add_option(
        "d",
        "detector",
        None,
        "The detector; either give as object or as (class, object_id)",
    )

    raw_parser.add_option(
        "mo",
        "mjdobs",
        "",
        "a measure for the start of the exposure, leave empty for value of tpl__start",
    )
    raw_parser.add_option(
        "ot", "obs__tplno", 0, "the number of the template in the observing block"
    )
    raw_parser.add_option(
        "te",
        "tpl__expno",
        target.tpl__expno.default,
        target.tpl__expno.__doc__,
    )
    raw_parser.add_option("ts", "tpl__start", "now", "the start of the template")

    def execute(self):
        """run the recipe"""
        args = self.init_arguments()
        for count in range(self.nr_raws):
            args["tpl__expno"] = self.tpl__expno + count
            # TODO tpl__start is used for filename, that should be mjdobs, which should be set per exposure
            # args["tpl__start"] = datetime.now().timestamp()
            obj = self.target(**args)
            obj.make()
            obj.save_fits()
            if self.commit:
                if not self.skip_store:
                    obj.store()
                obj.commit()
            self.process_result.append(obj.object_id.binary())

    def init_arguments(self):
        """return arguments needed for __init__ of target"""
        # default now, if given query for existing ?
        if self.tpl__start == "now":
            tpl__start = datetime.now()
        else:
            tpl__start = datetime.fromisoformat(self.tpl__start)
        if self.mjdobs:
            mjdobs = datetime.fromisoformat(self.mjdobs)
        else:
            mjdobs = tpl__start
        # possible parameters for MicadoImgCalDark(det__dit=1, det__ndit=1)
        if self.template:
            template = deserialize_object(self.template)
        else:
            template = self.template_default()
        # configure name, target
        if self.observing_block:
            observing_block = deserialize_object(self.observing_block)
        else:
            # TODO also set name?
            # TODO: Add proper default for 'instrument'.
            observing_block = ObservingBlock(
                target=self.observingblock_target,
                templates=[template],
                instrument="MICADO",
            )
        # possible parameters for generate_detectorarray(dark=dark_real, size=size, n_det=1)
        if self.detector:
            detectorarray = deserialize_object(self.detector)
        elif self.use_test_detectors:
            detectorarray = mini_default_detectorarray
        else:
            detectorarray = default_detectorarray
        return {
            "detector": detectorarray,
            "mjdobs": mjdobs.timestamp(),
            "obs__tplno": self.obs__tplno,
            "observing_block": observing_block,
            "template": template,
            "tpl__expno": self.tpl__expno,
            "tpl__start": tpl__start.timestamp(),
        }


class SimulateOptJob(SimulateRawJob):
    """job runs simulate Raw Opt data"""

    opt_parser = copy.deepcopy(SimulateRawJob.raw_parser)
    opt_parser.name = __name__

    opt_parser.add_option(
        "i", "instrument", None, "ESO acronym for the instrument used."
    )
    # TODO: filtname is currently ignored, because ScopeSIM only uses
    #       the filter specified in instrument.filter_wheel_1 and _23
    #       So filtname needs to be translated to the necessary FilterWheel
    #       settings.
    opt_parser.add_option("f", "filtname", "open", "Filter name.")
    opt_parser.add_option("ro", "relay_optics", None, "RelayOptics")

    def init_arguments(self):
        """return arguments needed for __init__ of target"""
        # Use given instrument or default?
        if self.instrument:
            instrument = deserialize_object(self.instrument)
        else:
            instrument = Instrument(
                name="MICADO",
                filter_wheel_1=FilterWheel(name="open"),
                filter_wheel_2=FilterWheel(name="Ks"),
                static_surfaces=SurfaceList(
                    realfilename="LIST_MICADO_mirrors_static.dat"
                ),
                ncpas_psf=NonCommonPathAberration(
                    realfilename="INST_MICADO_wavefront_error_budget.dat"
                ),
            )
        # Use given relay optics or default?
        if self.relay_optics:
            relay_optics = deserialize_object(self.relay_optics)
        else:
            relay_optics = RelayOptics(
                psf=FieldConstantPSF(
                    realfilename="PSF_SCAO_ConstPSF_0_5off.fits",
                ),
                surface_list=SurfaceList(
                    realfilename="LIST_RO_SCAO_mirrors.dat",
                ),
            )
        args = super().init_arguments()
        args["instrument"] = instrument
        args["filtname"] = self.filtname
        args["relay_optics"] = relay_optics
        return args


class SimulateSkyJob(SimulateOptJob):
    """job runs simulate Raw Sky data"""

    sky_parser = copy.deepcopy(SimulateOptJob.opt_parser)
    sky_parser.name = __name__

    sky_parser.add_option("a", "atmosphere", None, "The Atmosphere")
    sky_parser.add_option(
        "tc",
        "telescope",
        None,
        "ESO telescope code as defined in GEN-SPE-ESO-00000-0794",
    )

    def init_arguments(self):
        """return arguments needed for __init__ of target"""
        args = super().init_arguments()
        # Use given atmosphere?
        if self.atmosphere:
            args["atmosphere"] = deserialize_object(self.atmosphere)
        # Use given telescope?
        if self.telescope:
            args["telescope"] = deserialize_object(self.telescope)
        return args


class SimulateRawDarkJob(SimulateRawJob):
    """job runs simulate raw dark data"""

    # Target class to make
    target = Dark

    # default template class to use
    template_default = MicadoImgCalDark

    # observing block target name
    observingblock_target = "empty_sky"

    parser = copy.deepcopy(SimulateRawJob.raw_parser)
    parser.name = __name__


class SimulateRawFlatJob(SimulateOptJob):
    """job runs simulate raw flat data"""

    # Target class to make
    target = Flat

    # default template class to use
    template_default = MicadoImgCalFlat

    # observing block target name
    observingblock_target = "flatlamp"

    parser = copy.deepcopy(SimulateOptJob.opt_parser)
    parser.name = __name__


class SimulateRawScienceJob(SimulateSkyJob):
    """job runs simulate raw science data"""

    # Target class to make
    target = Science

    # default template class to use
    template_default = MicadoImgObs

    # observing block target name
    observingblock_target = "empty_sky"

    parser = copy.deepcopy(SimulateSkyJob.sky_parser)
    parser.name = __name__


class SimulateRawStdfieldJob(SimulateSkyJob):
    """job runs simulate raw science data"""

    # Target class to make
    target = Stdfield

    # default template class to use
    template_default = MicadoImgCalZeropoint

    # observing block target name
    observingblock_target = "cluster"

    parser = copy.deepcopy(SimulateSkyJob.sky_parser)
    parser.name = __name__


class SimulateRawIllumJob(SimulateSkyJob):
    """job runs simulate raw science data"""

    # Target class to make
    target = Illum

    # default template class to use
    template_default = MicadoImgTecIllum

    # observing block target name
    observingblock_target = "cluster"

    parser = copy.deepcopy(SimulateSkyJob.sky_parser)
    parser.name = __name__
