import glob
import importlib
from pathlib import Path

# Primarily import aweimports first to fill the classcache.
from micado.main.aweimports import Img, Cat, Tab
from micado.dataitems.Astrom import Astrom
from micado.dataitems.AstromRef import AstromRef
from micado.dataitems.Astromfield import Astromfield
from micado.dataitems.Background import Background
from micado.dataitems.Bkgfield import Bkgfield
from micado.dataitems.CalibHCI import CalibHCI
from micado.dataitems.Calibrated import Calibrated
from micado.dataitems.CalibratedCatalog import CalibratedCatalog
from micado.dataitems.Cam import Cam
from micado.dataitems.CoroBackground import CoroBackground
from micado.dataitems.CoroOffset import CoroOffset
from micado.dataitems.CubeHCI import CubeHCI
from micado.dataitems.Dark import Dark
from micado.dataitems.DepersistedBase import DepersistedBase
from micado.dataitems.DepersistedDark import DepersistedDark
from micado.dataitems.Detrend import Detrend
from micado.dataitems.DetrendAi import DetrendAi
from micado.dataitems.DetrendAstromfield import DetrendAstromfield
from micado.dataitems.DetrendBase import DetrendBase
from micado.dataitems.DetrendBkgfield import DetrendBkgfield
from micado.dataitems.DetrendCoroBackground import DetrendCoroBackground
from micado.dataitems.DetrendFinal import DetrendFinal
from micado.dataitems.DetrendFinalAi import DetrendFinalAi
from micado.dataitems.DetrendFinalAstromfield import DetrendFinalAstromfield
from micado.dataitems.DetrendFinalBase import DetrendFinalBase
from micado.dataitems.DetrendFinalHCIOffset import DetrendFinalHCIOffset
from micado.dataitems.DetrendFinalPupil import DetrendFinalPupil
from micado.dataitems.DetrendFinalStdfield import DetrendFinalStdfield
from micado.dataitems.DetrendFinalWaffle import DetrendFinalWaffle
from micado.dataitems.DetrendFinalWam import DetrendFinalWam
from micado.dataitems.DetrendHCIBase import DetrendHCIBase
from micado.dataitems.DetrendHCIOffset import DetrendHCIOffset
from micado.dataitems.DetrendIllum import DetrendIllum
from micado.dataitems.DetrendPupil import DetrendPupil
from micado.dataitems.DetrendStdfield import DetrendStdfield
from micado.dataitems.DetrendWaffle import DetrendWaffle
from micado.dataitems.DetrendWam import DetrendWam
from micado.dataitems.DistortionBase import DistortionBase
from micado.dataitems.DistortionElt import DistortionElt
from micado.dataitems.DistortionRefBase import DistortionRefBase
from micado.dataitems.DistortionWam import DistortionWam
from micado.dataitems.DpsicsOpt import DpsicsOpt
from micado.dataitems.DpsicsStaroffsetBase import DpsicsStaroffsetBase
from micado.dataitems.DpsicsStaroffsetOpt import DpsicsStaroffsetOpt
from micado.dataitems.DpsicsStaroffsetSci import DpsicsStaroffsetSci
from micado.dataitems.Flat import Flat
from micado.dataitems.Illum import Illum
from micado.dataitems.IllumCorr import IllumCorr
from micado.dataitems.MasterDark import MasterDark
from micado.dataitems.MasterFlat import MasterFlat
from micado.dataitems.Nonlin import Nonlin
from micado.dataitems.Nonlinearity import Nonlinearity
from micado.dataitems.Persistence import Persistence
from micado.dataitems.PhotomBase import PhotomBase
from micado.dataitems.PhotomRef import PhotomRef
from micado.dataitems.PhotomStdfield import PhotomStdfield
from micado.dataitems.Regrid import Regrid
from micado.dataitems.Science import Science
from micado.dataitems.ScienceAi import ScienceAi
from micado.dataitems.ScienceCoroPupil import ScienceCoroPupil
from micado.dataitems.ScienceCoroWaffle import ScienceCoroWaffle
from micado.dataitems.Stack import Stack
from micado.dataitems.StarcenterBase import StarcenterBase
from micado.dataitems.StarcenterPupil import StarcenterPupil
from micado.dataitems.StarcenterWaffle import StarcenterWaffle
from micado.dataitems.Staroffset import Staroffset
from micado.dataitems.Stdfield import Stdfield
from micado.dataitems.Wam import Wam
from micado.dataitems.WamRef import WamRef
from micado.main.DataItem import DataItem
from micado.main.Opt import Opt
from micado.main.Raw import Raw
from micado.main.Recipe import Recipe
from micado.main.Sky import Sky
from micado.recipes.ApplyJustPersistence import ApplyJustPersistence
from micado.recipes.McdDetDepersist import McdDetDepersist
from micado.recipes.McdImgCatalogableBase import McdImgCatalogableBase
from micado.main.commonwrapper import classcache

from micado.util.paths import PATH_DATAITEMS, PATH_TEMPLATE_PY


def get_dataitem_classes():
    """Return all processed dataitems."""
    # Import all dataitem classes.
    # TODO: Ensure that the ordering of the files has no effect on the figures
    #   As it is, the RecipeStep figures have different whitespace depending
    #   on the ordering here. The sorted() at least makes it deterministic...
    fns1 = sorted(glob.glob(str(PATH_DATAITEMS / "*.py")))
    fns2 = [Path(fn).parts[-1].replace(".py", "") for fn in fns1]
    fns3 = [fn for fn in fns2 if "extra" not in fn.lower() and "_" not in fn]
    mods = {fn: importlib.import_module(f"micado.dataitems.{fn}") for fn in fns3}
    classes = [getattr(mod, fn) for fn, mod in mods.items()]
    return [
        c
        for c in classes
        if issubclass(c, (Img, Tab, Cat))
        # TODO: Find a better way to include dataitems that share a recipe.
        and hasattr(c, "recipe")
        and c
        not in {
            # Only do normal Detrend.
            DetrendIllum,
            DetrendStdfield,
            DetrendWaffle,
            DetrendPupil,
            DetrendFinalStdfield,
            DetrendFinalWaffle,
            DetrendFinalPupil,
            DetrendWam,
            DetrendAstromfield,
            DpsicsStaroffsetBase,
            # TODO: Find better way to include/exclude DPSICS things.
            DpsicsStaroffsetOpt,
            DpsicsStaroffsetSci,
        }
    ]


def get_template_classes():
    """Import all template classes."""
    fns1 = glob.glob(str(PATH_TEMPLATE_PY / "M*.py"))
    fns1a = [fn for fn in fns1 if "Base" not in fn and "Extra" not in fn]
    fns2 = [Path(fn).parts[-1].replace(".py", "") for fn in fns1a]
    mods = {fn: importlib.import_module(f"micado.templates.{fn}") for fn in fns2}
    return [getattr(mod, fn) for fn, mod in mods.items()]


CLASSES_TEMPLATES = get_template_classes()

CLASSES_DATAITEMS = get_dataitem_classes()

CLASSES_RAW = [
    # Common with Spectroscopy.
    Dark,
    Flat,
    # Standard Imaging.
    Nonlin,
    Illum,
    Stdfield,
    Bkgfield,
    Science,
    # High Contrast Imaging.
    ScienceCoroWaffle,
    CoroOffset,
    ScienceCoroPupil,
    CoroBackground,
    ScienceAi,
    Wam,
    Cam,
    Astromfield,
]
# TODO: Create table for CLASSES_ICS.
CLASSES_ICS = [
    DpsicsOpt,
    ScienceCoroWaffle,
]
CLASSES_EXTERNAL = [
    AstromRef,
    PhotomRef,
    WamRef,
]
CLASSES_PRO = [
    Astrom,
    Background,
    MasterDark,
    MasterFlat,
    Detrend,
    DetrendFinal,
    IllumCorr,
    Persistence,
    PhotomBase,
    Stack,
    DetrendWaffle,
    DetrendPupil,
    StarcenterWaffle,
    StarcenterPupil,
]

CLASSES_STANDARD_IMAGING_SMALL = [
    DetrendFinalBase,
    Astrom,
    PhotomStdfield,
    Calibrated,
    Stack,
]

CLASSES_STANDARD_IMAGING = [
    MasterDark,
    MasterFlat,
    DetrendBase,
    Background,
] + CLASSES_STANDARD_IMAGING_SMALL

CLASSES_STANDARD_IMAGING_FINAL = [
    Calibrated,
    Stack,
]

CLASSES_STANDARD_IMAGING_ALL = CLASSES_STANDARD_IMAGING + [
    Detrend,
    DetrendStdfield,
    DetrendBkgfield,
    DetrendFinal,
    DetrendFinalStdfield,
    Dark,
    Flat,
    Science,
    Stdfield,
    Bkgfield,
]

CLASSES_ASTROMETRIC_IMAGING_SMALL = [
    DetrendFinalBase,
    DistortionBase,
    Astrom,
    PhotomStdfield,
    Calibrated,
    CalibratedCatalog,
    Regrid,
]

CLASSES_ASTROMETRIC_IMAGING = [
    MasterDark,
    MasterFlat,
    DetrendBase,
    Background,
] + CLASSES_ASTROMETRIC_IMAGING_SMALL

CLASSES_ASTROMETRIC_IMAGING_FINAL = [
    CalibratedCatalog,
    Regrid,
]

CLASSES_ASTROMETRIC_IMAGING_ALL = CLASSES_ASTROMETRIC_IMAGING + [
    DistortionWam,
    DistortionElt,
    DetrendAi,
    DetrendStdfield,
    DetrendBkgfield,
    DetrendWam,
    DetrendAstromfield,
    DetrendFinalAi,
    DetrendFinalStdfield,
    DetrendFinalWam,
    DetrendFinalAstromfield,
    Dark,
    Flat,
    ScienceAi,
    Stdfield,
    Bkgfield,
    Wam,
    Astromfield,
    WamRef,
    AstromRef,
]

CLASSES_HIGH_CONTRAST_IMAGING_FOCAL = [
    MasterDark,
    MasterFlat,
    DetrendHCIBase,
    Background,
    DetrendFinalBase,
    Staroffset,
    StarcenterWaffle,
    CalibHCI,
    CubeHCI,
]

CLASSES_HIGH_CONTRAST_IMAGING_PUPIL = [
    MasterDark,
    MasterFlat,
    DetrendHCIBase,
    Background,
    DetrendFinalBase,
    Staroffset,
    StarcenterPupil,
    CalibHCI,
    CubeHCI,
]

CLASSES_HIGH_CONTRAST_IMAGING_SMALL = [
    DetrendFinalBase,
    Staroffset,
    StarcenterWaffle,
    StarcenterPupil,
    CalibHCI,
    CubeHCI,
]

CLASSES_HIGH_CONTRAST_IMAGING = [
    MasterDark,
    MasterFlat,
    DetrendHCIBase,
    Background,
] + CLASSES_HIGH_CONTRAST_IMAGING_SMALL

# TODO: Generate CLASSES_HIGH_CONTRAST_IMAGING automatically
assert set(
    CLASSES_HIGH_CONTRAST_IMAGING_FOCAL + CLASSES_HIGH_CONTRAST_IMAGING_PUPIL
) == set(CLASSES_HIGH_CONTRAST_IMAGING)

CLASSES_HIGH_CONTRAST_IMAGING_FOCAL_ALL = CLASSES_HIGH_CONTRAST_IMAGING_FOCAL + [
    Dark,
    Flat,
    ScienceCoroWaffle,
    CoroOffset,
    CoroBackground,
    DetrendWaffle,
    DetrendHCIOffset,
    DetrendFinalHCIOffset,
    DetrendCoroBackground,
    DetrendFinalWaffle,
]
CLASSES_HIGH_CONTRAST_IMAGING_FOCAL_ALL.remove(DetrendHCIBase)

CLASSES_HIGH_CONTRAST_IMAGING_PUPIL_ALL = CLASSES_HIGH_CONTRAST_IMAGING_PUPIL + [
    Dark,
    Flat,
    ScienceCoroPupil,
    CoroOffset,
    DetrendPupil,
    DetrendHCIOffset,
    DetrendFinalHCIOffset,
    DetrendFinalPupil,
]
CLASSES_HIGH_CONTRAST_IMAGING_PUPIL_ALL.remove(DetrendHCIBase)

CLASSES_HIGH_CONTRAST_IMAGING_ALL = CLASSES_HIGH_CONTRAST_IMAGING + [
    Dark,
    Flat,
    ScienceCoroPupil,
    ScienceCoroWaffle,
    CoroOffset,
    CoroBackground,
    DetrendCoroBackground,
    DetrendWaffle,
    DetrendPupil,
    DetrendHCIOffset,
    DetrendFinalHCIOffset,
    DetrendFinalWaffle,
    DetrendFinalPupil,
]
CLASSES_HIGH_CONTRAST_IMAGING_ALL.remove(DetrendHCIBase)

# TODO: Generate CLASSES_HIGH_CONTRAST_IMAGING_ALL automatically
assert set(
    CLASSES_HIGH_CONTRAST_IMAGING_FOCAL_ALL + CLASSES_HIGH_CONTRAST_IMAGING_PUPIL_ALL
) == set(CLASSES_HIGH_CONTRAST_IMAGING_ALL)

CLASSES_QACITS = [
    DpsicsStaroffsetOpt,
    DpsicsStaroffsetSci,
]

CLASSES_QACITS_ALL = CLASSES_QACITS + [
    DpsicsOpt,
    ScienceCoroWaffle,
]

CLASSES_UPTO_REDUCED = [
    MasterDark,
    MasterFlat,
    DetrendBase,
    Background,
    DetrendFinalBase,
]

CLASSES_UPTO_REDUCED_ALL = (
    CLASSES_STANDARD_IMAGING_ALL
    + CLASSES_ASTROMETRIC_IMAGING_ALL
    + CLASSES_HIGH_CONTRAST_IMAGING_ALL
)

CLASSES_UPTO_REDUCED_FINAL = []

CLASSES_CALIBRATION_PLAN = [
    IllumCorr,
    MasterDark,
    MasterFlat,
    Nonlinearity,
    # TODO: List PhotomBase here and fetch -StdField and -Science?
    PhotomStdfield,
    Staroffset,
    DistortionElt,
    DistortionWam,
]

# These should have their own pipelines.
CLASSES_WITHOUT_PIPELINE = [
    Persistence,
    Nonlinearity,
    IllumCorr,
    DistortionElt,
    # TODO: Remove Depersisted once removed from CPL code.
    DepersistedBase,
    DepersistedDark,
]

CLASSES_INTERMEDIATE_DRL_FUNCTION = [
    Img,
    DetrendFinal,
    PhotomBase,
    StarcenterBase,
    DistortionRefBase,
    DistortionBase,
]

RECIPES_ALL = [
    v.aclass
    for v in classcache.values()
    if issubclass(v.aclass, Recipe) and v.aclass not in [Recipe, McdImgCatalogableBase]
    # TODO: Find better way to include/exclude DPSICS libraries
    and not issubclass(v.aclass, DpsicsStaroffsetBase.recipe.prop_type)
]


DATAITEMS_ALL = [
    v.aclass for v in classcache.values() if issubclass(v.aclass, DataItem)
]

# Reference data items.
CLASSES_REF = [AstromRef, PhotomRef, WamRef]

# Classes that should not show up in the DRLD, CP, or DRLVT.
CLASSES_HIDE_FROM_ESO_DOCUMENTATION = [
    # Data items
    DepersistedDark,
    DpsicsStaroffsetSci,
    DepersistedBase,
    DepersistedDark,
    # Recipes / DRL Functions
    McdDetDepersist,
    ApplyJustPersistence,
    # Base classes
    Raw,
    Opt,
    Sky,
] + [c for c in DATAITEMS_ALL if c.name_file_doc().lower().endswith("base")]
