# -*- coding: utf-8 -*-
"""Create full dictionary of ESO keywords."""

import glob
import json
import pathlib
import re
from collections import OrderedDict

from astropy.io import fits


# TODO: Perhaps make a proper (data)class of these keywords?
def get_keyword_dict():
    """Get dictionary of keywords from json files."""
    dir_json = (
        pathlib.Path(__file__).resolve().parent.parent.parent / "datainterface" / "yaml"
    )
    globs = str(dir_json / "*.flat.json")
    fns = glob.glob(globs)
    kds = [kd for fn in fns for kd in json.load(open(fn))]
    kkeys = [kd["name"] for kd in kds]
    keys_extra = [k for k in kkeys if kkeys.count(k) > 1]
    assert keys_extra == [], keys_extra
    kds2 = OrderedDict([(kd["name"], kd) for kd in kds])
    kds3 = OrderedDict(sorted(kds2.items()))

    # For E.g. INS.FILTi.NAME, also add INS.FILT1.NAME and INS.FILT2.NAME etc.
    # TODO: Perhaps also for other lower case letters?
    for k, v in list(kds3.items()):
        if "i" in k:
            for s in "0123456789":
                kds3[k.replace("i", s)] = v

    return kds3


def verify_fitsheaders(filename: str):
    """Verify the headers in a FITS file.

    Args:
        filename: Name of a FITS file to verify the headers of.
    """
    with fits.open(filename) as ffits:
        print(ffits[0].header.tostring(sep="\n"))
        print(ffits[1].header.tostring(sep="\n"))

        errors = []

        key_src = "FNSRC0"
        # The source is allowed to be 'scopesim_templates.micado.empty_sky()'
        # but not 'scopesim_templates.micado.cluster(mass=5)'
        if key_src in ffits[0].header:
            value_src = ffits[0].header[key_src]
            if "=" in value_src:
                errors.append(f"{key_src} should not have an =-sign: {value_src}")

        for hk in ffits[0].header:
            if hk.startswith("SIM SRC") or hk.startswith("FNSRC"):
                # Internal ScopeSIM header.
                continue
            hk2 = hk[4:] if hk.startswith("ESO ") else hk
            hk2 = hk[5:] if hk.startswith("WISE ") else hk2
            hk3 = re.sub("[\\d]", "i", hk2)
            # Replace single underscore, e.g. for DET_DIT which is copied from
            # the templates also to the RAW data.
            hk4 = hk3.replace("_", " ")
            # Raw data usually has a detector array in the first header, that
            # is, DET.DET2.XHW keywords etc. This is a mistake, these headers
            # should only be described in the extension headers. That is,
            # extension 2 should have just DET.XHW and there should be no
            # DET.DET2.XHW.
            # TODO: Fix the above detector array in primary header bug.
            hk5 = hk3.replace("DETi ", "")
            # The 'i' is not necessarily included in the DID if there can
            # also only be one. E.g. FILT. There should never be
            # just a FILT, but it doesn't make sense to define it as FILTi.
            hk6 = hk3.replace("i", "")
            if (
                hk2 not in esokeywords
                and hk3 not in esokeywords
                and hk4 not in esokeywords
                and hk5 not in esokeywords
                and hk6 not in esokeywords
            ):
                errors.append(f"Warning! {hk2} not in esokeywords")

        assert not errors, "\n".join(errors)


esokeywords = get_keyword_dict()
